#
# Copyright (c) 2020, Hyve Design Solutions Corporation.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are 
# met:
#
# 1. Redistributions of source code must retain the above copyright 
#    notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright 
#    notice, this list of conditions and the following disclaimer in the 
#    documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of Hyve Design Solutions Corporation nor the names 
#    of its contributors may be used to endorse or promote products 
#    derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY HYVE DESIGN SOLUTIONS CORPORATION AND 
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, 
# BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND 
# FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 
# HYVE DESIGN SOLUTIONS CORPORATION OR CONTRIBUTORS BE LIABLE FOR ANY 
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL 
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS 
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
# STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING 
# IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
# POSSIBILITY OF SUCH DAMAGE.
#
from collections import OrderedDict
from .. mesg.ipmi_chassis import *
from . _consts import TupleExt
from . _common import str2int, do_command
from .. util.exception import PyCmdsExcept, PyCmdsArgsExcept

def _chassis_status(self, argv):
    t1 = self.intf.issue_cmd(GetChsStat)
    d1 = OrderedDict()

    get_on = lambda val, mask: 'on' if val & mask else 'off'
    get_true = lambda val, mask: 'true' if val & mask else 'false'
    get_act = lambda val, mask: 'active' if val & mask else 'inactive'

    # Current Power State
    d1['System Power'] = get_on(t1[0], 1)
    d1['Power Overload'] = get_true(t1[0], 2)
    d1['Power Interlock'] = get_act(t1[0], 4)
    d1['Main Power Fault'] = get_true(t1[0], 8)
    d1['Power Control Fault'] = get_true(t1[0], 16)
    d1['Power Restore Policy'] = ('always-off', 'last state', 'always-on', 'unknown')[(t1[0] & 0x60) >> 5]

    # Last Power Event
    events = []
    event_names = ('AC failed', 'Power overload', 'Power interlock', 'Power fault', 'Power on by IPMI')
    e = t1[1] & 0x1f

    idx = 0
    while e:
        if e & 1: events.append(event_names[idx])
        e >>= 1
        idx += 1

    d1['Last Power Event'] = ', '.join(events) if events else ''

    # Misc Chassis Status
    d1['Chassis Intrusion'] = get_act(t1[2], 1)
    d1['Front-Panel Lockout'] = get_act(t1[2], 2)
    d1['Drive Fault'] = get_true(t1[2], 4)
    d1['Cooling/Fan Fault'] = get_true(t1[2], 8)
    if t1[2] & 64:
        d1['Chassis Identify'] = ('off', 'temporary on', 'indefinite on', 'reserved')[(t1[2] & 0x30) >> 4]

    # FP Button Cap and status
    if len(t1) >= 4:
        d1['Power Button Disabled'] = get_true(t1[3], 1)
        d1['Reset Button Disabled'] = get_true(t1[3], 2)
        d1['Diag Button Disabled'] = get_true(t1[3], 4)
        d1['Sleep Button Disabled'] = get_true(t1[3], 8)

        get_allow = lambda mask: 'allowed' if t1[3] & mask else 'not allowed'
        d1['Power Button Disable'] = get_allow(16)
        d1['Reset Button Disable'] = get_allow(32)
        d1['Diag Button Disable'] = get_allow(64)
        d1['Sleep Button Disable'] = get_allow(128)

    # Output
    for key, val in d1.items():
        self.print('{0:24}: {1}'.format(key, val))

def _chassis_power(self, argv):
    from . _power import do_power
    do_power(self, argv[1:])

def _chassis_policy(self, argv):
    if len(argv) < 2:
        raise PyCmdsArgsExcept(1, 1)

    opts = TupleExt(('always-off', 'previous', 'always-on', 'list'))
    opt = opts.get_index(argv[1])
    if opt < 0:
        raise PyCmdsArgsExcept(2, 1, argv[1])

    poc, = self.intf.issue_cmd(PowerRestore, opt)

    if opt == 3:    # list (no change)
        poc_support = []
        poc &= 7
        idx = 0
        while poc:
            if poc & 1: poc_support.append(opts[idx])
            poc >>= 1
            idx += 1
        ret = ' '.join(poc_support) if poc_support else '(none)'
        self.print('Supported chassis power policy: ' + ret)
    else:        
        self.print('Set chassis power restore policy to ' + opts[opt])

def _chassis_restart_cause(self, argv):
    cause, _ = self.intf.issue_cmd(GetSysRestartCause)
    rcs = ('unknown', 'Chassis Control command', 'reset via pushbutton',
           'power-up via power pushbutton', 'Watchdog expiration', 'OEM',
           'automatic power-up on AC being applied due to \'always restore\' power restore policy',
           'automatic power-up on AC being applied due to \'restore previous power state\' power restore policy',
           'reset via PEF', 'power-cycle via PEF', 'soft reset', 'power-up via RTC wakeup')
    
    if cause >= len(rcs):  cause = 0
    self.print('System restart cause: ' + rcs[cause])

def _chassis_poh(self, argv):
    mins, counter = self.intf.issue_cmd(GetPoh)
    mins_all = mins * counter
    
    h = int(mins_all / 60)
    m = mins_all - h * 60 

    d = int(h / 24)
    h -= d * 24

    self.print('POH Counter: {0} days, {1} hours, {2} minutes'.format(d, h, m))

def _chassis_identify(self, argv):
    interval = 15
    force = 0    

    if len(argv) >= 2:
        if argv[1] == 'on':
            force = 1
            ret = 'indefinitely on'
        elif argv[1] == 'off':
            interval = 0
        else:
            int_req = str2int(argv[1])
            if int_req > 0 and int_req < 256:
                interval = int_req
            ret = '{0} seconds'.format(interval)
    else:
        ret = 'default ({0} seconds)'.format(interval)

    self.intf.issue_cmd(ChsIdentify, interval, force)

    CI = 'Chassis identify'
    if interval == 0:
        self.print(CI + ' is off')
    else:
        self.print(CI + ' interval: ' + ret)

def _chassis_bootdev(self, argv):
    if len(argv) < 2: 
        raise PyCmdsArgsExcept(1, 2)
    
    devs = TupleExt(('none', 'pxe', 'disk', 'safe', 'diag', 'cdrom', 'bios'))
    if argv[1] == 'floppy':
        dev = 15
    else:
        dev = devs.get_index(argv[1])

    if dev == -1:
        raise PyCmdsArgsExcept(2, 2, argv[1])

    if len(argv) >= 3:
        if argv[2] == 'clear-cmos':
            dev |= 0x80
        else:
            raise PyCmdsArgsExcept(2, 2, argv[2])

    # data1 = 1010 0000b, this will change the config to EFI
    data = struct.pack('BBBBB', 0xa0, dev, 0, 0, 0)
    
    self.intf.issue_cmd(SetBootOpts, 5, data)

def help_chassis(self, argv=None, context=0):
    if context == 0:
        self.print('Chassis Commands:')
        self.print('    status')        
        self.print('    power <status|on|off|cycle|reset|diag|soft>')
    if context in (0, 1):
        self.print('    policy <list|always-on|previous|always-off>')
    if context == 0:
        self.print('    restart_cause')
        self.print('    poh')
        self.print('    identify [<on|off|1-255>]')
    if context in (0, 2):
        self.print('    bootdev <none|pxe|disk|safe|diag|cdrom|bios|floppy> [clear-cmos]')

CHASSIS_CMDS = {
    'status': _chassis_status,
    'power': _chassis_power,
    'policy': _chassis_policy,
    'restart_cause': _chassis_restart_cause,
    'poh': _chassis_poh,
    'identify': _chassis_identify,
    'bootdev': _chassis_bootdev,
    'help': help_chassis,
}

def do_chassis(self, argv):
    do_command(self, argv, help_chassis, CHASSIS_CMDS)



#
# Copyright (c) 2020, Hyve Design Solutions Corporation.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are 
# met:
#
# 1. Redistributions of source code must retain the above copyright 
#    notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright 
#    notice, this list of conditions and the following disclaimer in the 
#    documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of Hyve Design Solutions Corporation nor the names 
#    of its contributors may be used to endorse or promote products 
#    derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY HYVE DESIGN SOLUTIONS CORPORATION AND 
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, 
# BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND 
# FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 
# HYVE DESIGN SOLUTIONS CORPORATION OR CONTRIBUTORS BE LIABLE FOR ANY 
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL 
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS 
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
# STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING 
# IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
# POSSIBILITY OF SUCH DAMAGE.
#
import struct
from .. mesg.ipmi_app import GetDeviceGUID, GetDeviceID, GetSelfTest, ColdReset, \
                             GetEnables, SetEnables, GetSysInfo, SetSysInfo, \
                             ResetWatchdog, SetWatchdog, GetWatchdog
from . _common import bcd2str, do_command
from . _consts import TupleExt
from .. util.exception import PyCmdsArgsExcept

def _mc_reset(self, argv):
    self.intf.issue_cmd(ColdReset)

def _mc_guid(self, argv):
    guid, = self.intf.issue_cmd(GetDeviceGUID)
    self.print('Device GUID:', guid.hex())

def _mc_info(self, argv):
    t1 = self.intf.issue_cmd(GetDeviceID)
    self.print('Device ID                 : {0}'.format(t1[0]))
    self.print('Device Revision           : {0}'.format(t1[1] & 0x0f))
    self.print('Firmware Revision         : {0}.{1}'.format(t1[2] & 0x7f, bcd2str(t1[3])))
    self.print('IPMI Version              : {0}'.format(bcd2str(t1[4], True)))
    self.print('Manufacturer ID           : 0x{0:06x}'.format(int.from_bytes(t1[6], 'little')))
    self.print('Product ID                : {0}'.format(t1[7]))    

    dev_avail = 'yes' if t1[2] & 0x80 == 0 else 'no'
    self.print('Device Available          : {0}'.format(dev_avail))

    dev_sdr = 'no' if t1[1] & 0x80 == 0 else 'yes'
    self.print('Provides Device SDRs      : {0}'.format(dev_sdr))

    self.print('Additional Device Support :')
    devs = ('Sensor Device', 'SDR Repository Device', 'SEL Device',
            'FRU Inventory Device', 'IPMB Event Receiver',
            'IPMB Event Generator', 'Bridge', 'Chassis Device',)
    dev_sup = t1[5]

    for i in range(8):
        if dev_sup & 1:  self.print('    ' +  devs[i])
        dev_sup >>= 1

    self.print('Aux Firmware Rev Info     :')
    for i in t1[8]:
        self.print('    0x{0:02x}'.format(i))

def _mc_selftest(self, argv):
    rsp, err = self.intf.issue_cmd(GetSelfTest)

    msg = {0x55: 'passed', 0x56: 'not implemented', 
           0x57: 'corrupted or inaccessible data or devices',
           0x58: 'fatal hardware error', 0xff: 'reserved'}

    if rsp in msg.keys():
        self.print('Selftest: {0} ({1:02X}h)'.format(msg[rsp], rsp))
        if rsp == 0x57:
            errs = ('controller operational firmware corrupted',
                    'controller update \'boot block\' firmware corrupted',
                    'Internal Use Area of BMC FRU corrupted',
                    'SDR Repository empty',
                    'IPMB signal lines do not respond',
                    'cannot access BMC FRU device',
                    'cannot access SDR Repository',
                    'cannot access SEL device',)
            for i in range(8):
                if err & 1:  self.print('    ' + errs[i])
                err >>= 1                    
    else:
        self.print('Selftest: {0:02X}h'.format(rsp))

def _mc_getenables(self, argv):
    rsp, = self.intf.issue_cmd(GetEnables)
    enables = ('Receive Message Queue Interrupt', 'Event Message Buffer Full Interrupt',
               'Event Message Buffer', 'System Event Logging', 'reserved', 
               'OEM 0', 'OEM 1', 'OEM 2',)
    for i in range(8):
        if i != 4:  
            self.print('{0:36}'.format(enables[i]), end=' : ')
            if rsp & 1: self.print('enabled')
            else:   self.print('disabled')
        rsp >>= 1

def _mc_setenables(self, argv):
    if len(argv) < 3:
        raise PyCmdsArgsExcept(1, 2)
    if argv[2] not in ('on', 'off'):
        raise PyCmdsArgsExcept(2, 2, argv[2])
        
    cmds = TupleExt(('recv_msg_intr', 'event_msg_intr', 'event_msg', 'system_event_log',
                     'reserved', 'oem0', 'oem1', 'oem2',))
    idx = cmds.get_index(argv[1])
    if idx < 0: 
        raise PyCmdsArgsExcept(2, 2, argv[1])

    enables, = self.intf.issue_cmd(GetEnables)
    val = idx ** 2
    if argv[2] == 'on':  enables |= val
    else:  enables &= ~val

    self.intf.issue_cmd(SetEnables, enables)

def _sysinfo_opts(self, argv, chk):
    idx = 0
    if len(argv) < chk:
        raise PyCmdsArgsExcept(1, 3)

    params = TupleExt(('none', 'system_fw_version', 'system_name', 'primary_os_name', 
                       'os_name',))        
    idx = params.get_index(argv[1], 0)
    if idx == 0:
        raise PyCmdsArgsExcept(2, 3, argv[1])

    return idx

def _mc_getsysinfo(self, argv):
    idx = _sysinfo_opts(self, argv, 2)

    rsp = self.intf.issue_cmd(GetSysInfo, idx, 0, 0)
    if not rsp:
        self.print('Option {0} is not supported by BMC.'.format(argv[1]))
        return

    ret = rsp[2:].decode('latin_1')
    ret_len = rsp[1] - 14
    sel = 1

    while ret_len > 0:
        rsp = self.intf.issue_cmd(GetSysInfo, idx, sel, 0)
        if not rsp: break
        ret += rsp.decode('latin_1')
        ret_len -= 16
        if sel == 255:  break
        sel += 1

    self.print(argv[1] + ': ' + ret.strip())

def _mc_setsysinfo(self, argv):
    idx = _sysinfo_opts(self, argv, 3)

    val = argv[2].encode('latin_1')
    val_len = len(val)
    if val_len > 255:   
        val = val[:255]
        val_len = 255
    
    data = struct.pack('BBB', 0, 0, val_len) + val[:14]
    val_len -= 14
    if val_len > 0: val = val[14:]
    elif val_len < 0: data += b'\0' * (-val_len)

    cc = self.intf.issue_cmd(SetSysInfo, idx, data)
    if cc != 0:
        if cc == 0x80:
            self.print('Option {0} is not supported by BMC.'.format(argv[1]))
        elif cc == 0x82:
            self.print('Option {0} is read-only.'.format(argv[1]))
        else:
            self.print('Error: completion code = {0:02X}h.'.format(cc))
        return

    sel = 1
    while val_len > 0:
        data = struct.pack('B', sel) + val[:16]
        val_len -= 16
        if val_len > 0: val = val[16:]
        elif val_len < 0: data += b'\0' * (-val_len)
        cc = self.intf.issue_cmd(SetSysInfo, idx, data)
        if cc != 0: return
        sel += 1

def _mc_watchdog(self, argv):
    if len(argv) < 2:
        raise PyCmdsArgsExcept(1, 1)
        
    if argv[1] not in ('get', 'reset'):
        raise PyCmdsArgsExcept(2, 1, argv[1])

    if argv[1] == 'get':
        (timer, act, pre_int, exp, init_count, 
            present_count) = self.intf.issue_cmd(GetWatchdog)

        timers = TupleExt(('reserved', 'BIOS FRB2', 'BIOS/POST', 'OS Load', 'SMS/OS', 'OEM'))
        timer_str = timers.get(timer & 7, 'reserved')
        timer_stat = ('Stopped', 'Started')[(timer & 0x40) >> 6]
        timer_log = ('On', 'Off')[(timer & 0x80) >> 7]

        acts = TupleExt(('No action', 'Hard Reset', 'Power Down', 'Power Cycle'))
        act_str = acts.get(act, 'reserved')
        ints = TupleExt(('None', 'SMI', 'NMI / Diagnostic Interrupt', 'Messaging Interrupt'))
        int_str = ints.get((act & 0x70) >>  4, 'reserved')

        exp >>= 1
        exp_flags = []
        for i in range(1, 6):
            if exp & 1:  exp_flags.append(timers[i])
            exp >>= 1

        self.print('Watchdog Timer Use       : {0} ({1:02X}h)'.format(timer_str, timer & 7))
        self.print('Watchdog Timer Is        : {0}'.format(timer_stat))
        self.print('Watchdog Timer Logging   : {0}'.format(timer_log))
        self.print('Watchdog Timer Action    : {0} ({1:02X}h)'.format(act_str, act & 7))
        self.print('Pre-timeout Interrupt    : {0}'.format(int_str))
        self.print('Pre-timeout Interval     : {0} seconds'.format(pre_int))
        
        if not exp_flags:
            self.print('Timer Expiration Flags   : none')
        else:
            self.print('Timer Expiration Flags   : [{0}]'.format(exp_flags[0]))
            for i in exp_flags[1:]:
                self.print(' ' * 27 + '[{0}]'.format(i))

        self.print('Initial Countdown        : {0} seconds'.format(init_count / 10))
        self.print('Present Countdown        : {0} seconds'.format(present_count / 10))

    else:   # reset    
        cc = self.intf.issue_cmd(ResetWatchdog)
        if cc == 0:
            self.print('IPMI Watchdog has been successfully reset.')
        elif cc == 0x80:
            self.print('Failed to reset IPMI Watchdog.  Attempted to start un-initialized watchdog.')
        else:
            self.print('Failed to reset IPMI Watchdog.  CC={0:2X}h.'.format(cc))

def help_mc(self, argv=None, context=0):
    if context == 0:
        self.print('MC Commands:')
        self.print('    help')
        self.print('    reset')
        self.print('    guid')
        self.print('    info')
        self.print('    selftest')
    if context in (0, 1):
        self.print('    watchdog <get|reset>')
    if context in (0, 2):
        self.print('    getenables')
        self.print('    setenables <option> <on|off>')
        self.print('     * recv_msg_intr')
        self.print('     * event_msg_intr')
        self.print('     * event_msg')
        self.print('     * system_event_log')
        self.print('     * oem0')
        self.print('     * oem1')
        self.print('     * oem2')
    if context in (0, 3):
        self.print('    getsysinfo <option>')
        self.print('    setsysinfo <option> <value>')
        self.print('     * system_fw_version')
        self.print('     * system_name')
        self.print('     * primary_os_name')
        self.print('     * os_name')

MC_CMDS = {
    'reset': _mc_reset,
    'guid': _mc_guid,
    'info': _mc_info,
    'watchdog': _mc_watchdog, 
    'selftest': _mc_selftest,
    'getenables': _mc_getenables,
    'setenables': _mc_setenables,
    'getsysinfo': _mc_getsysinfo,
    'setsysinfo': _mc_setsysinfo,
    'help': help_mc,  
}

def do_mc(self, argv):
    do_command(self, argv, help_mc, MC_CMDS)



#
# Copyright (c) 2020, Hyve Design Solutions Corporation.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are 
# met:
#
# 1. Redistributions of source code must retain the above copyright 
#    notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright 
#    notice, this list of conditions and the following disclaimer in the 
#    documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of Hyve Design Solutions Corporation nor the names 
#    of its contributors may be used to endorse or promote products 
#    derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY HYVE DESIGN SOLUTIONS CORPORATION AND 
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, 
# BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND 
# FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 
# HYVE DESIGN SOLUTIONS CORPORATION OR CONTRIBUTORS BE LIABLE FOR ANY 
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL 
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS 
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
# STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING 
# IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
# POSSIBILITY OF SUCH DAMAGE.
#
from .. util.exception import PyMesgExcept, PyMesgCCExcept
import struct

__all__ = [
    'IPMI_Message',
    'IPMI_Raw',
    'ipmi_app',
    'ipmi_chassis',
    'ipmi_se',
    'ipmi_storage',
    'ipmi_transport',
]

class IPMI_Message:
    def __init__(self, netfn, cmd, req_data=None, lun=0):
        self.netfn = netfn
        self.cmd = cmd
        self.req_data = req_data
        self.lun = lun
        self.payload_type = 0

    @staticmethod
    def dump_tuple(t1):
        if type(t1) is not tuple:
            return ''

        data_str = ', '.join(('{0:x}'.format(i) if type(i) is not bytes  
            else (' '.join('{0:02x}'.format(j) for j in i)) for i in t1))

        return '(' + data_str + ')'
    
    def unpack(self, rsp, fmt=None):
        # rsp = (netfn, cmd, cc, rsp_data)
        if rsp[0] != self.netfn + 1:
            raise PyMesgExcept('Invalid NetFn {0:02x}h in the response.'
                               .format(rsp[0]))

        if rsp[1] != self.cmd:
            raise PyMesgExcept('Invalid CMD {0:02x}h in the response.'
                               .format(rsp[1]))

        cc, rsp_data = rsp[2:] 
        if isinstance(self, IPMI_Raw):
            list1 = [cc]
            if rsp_data is not None:
                list1 += list(rsp_data)
            return list1

        if cc != 0:
            raise PyMesgCCExcept(self.netfn, self.cmd, cc)

        if fmt is None: return rsp_data     # do not unpack the response

        if rsp_data is None:    # no response data, but fmt is not None
            raise PyMesgExcept('Unexpected empty response data: NetFn={0:02X}h, CMD={1:02X}h.  Expected {2}.'
                               .format(self.netfn, self.cmd, struct.calcsize(fmt)))

        if struct.calcsize(fmt) != len(rsp_data):
            raise PyMesgExcept('Invalid response data length: NetFn={0:02X}h, CMD={1:02X}h.  Expected {2}, but returned {3}.'
                               .format(self.netfn, self.cmd, struct.calcsize(fmt), len(rsp_data)))

        return struct.unpack(fmt, rsp_data) # has response data

class IPMI_Raw(IPMI_Message):
    def __init__(self, req, lun=0):
        # [netfn, cmd, req_data]
        req_data = None
        if len(req) > 2:
            req_data = bytes(req[2:])            
        super(IPMI_Raw, self).__init__(req[0], req[1], req_data, lun)

    def unpack(self, rsp):
        # rsp = (netfn, cmd, cc, rsp_data)
        return super(IPMI_Raw, self).unpack(rsp)



#
# Copyright (c) 2020, Hyve Design Solutions Corporation.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are 
# met:
#
# 1. Redistributions of source code must retain the above copyright 
#    notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright 
#    notice, this list of conditions and the following disclaimer in the 
#    documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of Hyve Design Solutions Corporation nor the names 
#    of its contributors may be used to endorse or promote products 
#    derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY HYVE DESIGN SOLUTIONS CORPORATION AND 
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, 
# BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND 
# FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 
# HYVE DESIGN SOLUTIONS CORPORATION OR CONTRIBUTORS BE LIABLE FOR ANY 
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL 
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS 
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
# STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING 
# IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
# POSSIBILITY OF SUCH DAMAGE.
#
class TupleExt(tuple):
    def get(self, idx, default):
        if idx >= len(self):
            return default
        else:
            return self[idx]

    def get_index(self, val, default=-1):
        try:
            idx = super(TupleExt, self).index(val)
        except:
            idx = default
        finally:
            return idx

RATE_UNITS = TupleExt((
    'none', 'per us', 'per ms', 'per s', 'per minute',
    'per hour', 'per day',
))

BASE_UNITS = TupleExt((
    'unspecified', 'degrees C', 'degrees F', 'degrees K', 'Volts',
    'Amps', 'Watts', 'Joules', 'Coulombs', 'VA',
    'Nits', 'lumen', 'lux', 'Candela', 'kPa',
    'PSI', 'Newton', 'CFM', 'RPM', 'Hz',
    'microsecond', 'millisecond', 'second', 'minute', 'hour',
    'day', 'week', 'mil', 'inches', 'feet',
    'cu in', 'cu feet', 'mm', 'cm', 'm',
    'cu cm', 'cu m', 'liters', 'fluid ounce', 'radians',
    'steradians', 'revolutions', 'cycles', 'gravities', 'ounce',
    'pound', 'ft-lb', 'oz-in', 'gauss', 'gilberts',
    'henry', 'millihenry', 'farad', 'microfarad', 'ohms',
    'siemens', 'mole', 'becquerel', 'PPM', 'reserved',
    'Decibels', 'DbA', 'DbC', 'gray', 'sievert',
    'color temp deg K', 'bit', 'kilobit', 'megabit', 'gigabit',
    'byte', 'kilobyte', 'megabyte', 'gigabyte', 'word',
    'dword', 'qword', 'line', 'hit', 'miss', 'retry',
    'reset', 'overflow', 'underrun', 'collision', 'packets',
    'messages', 'characters', 'error', 'correctable error', 'uncorrectable error',
    'fatal error', 'grams',
))

SENSOR_TYPES = TupleExt((
    'reserved', 'Temperature', 'Voltage', 'Current', 'Fan',
    'Physical Security', 'Platform Security', 'Processor', 'Power Supply', 'Power Unit',
    'Cooling Device', 'Other', 'Memory', 'Drive Slot / Bay', 'POST Memory Resize',
    'System Firmwares', 'Event Logging Disabled', 'Watchdog1', 'System Event', 'Critical Interrupt',
    'Button', 'Module / Board', 'Microcontroller', 'Add-in Card', 'Chassis',
    'Chip Set', 'Other FRU', 'Cable / Interconnect', 'Terminator', 'System Boot Initiated',
    'Boot Error', 'OS Boot', 'OS Critical Stop', 'Slot / Connector', 'System ACPI Power State',
    'Watchdog2', 'Platform Alert', 'Entity Presence', 'Monitor ASIC', 'LAN',
    'Management Subsys Health', 'Battery', 'Session Audit', 'Version Change', 'FRU State',
))

ENTITY_ID_CODES = TupleExt((
    'Unspecified',
    'Other',
    'Unknown',
    'Processor',
    'Disk or Disk Bay',
    'Peripheral Bay',
    'System Management Module',
    'System Board',
    'Memory Module',
    'Processor Module',
    'Power Supply',
    'Add-in Card',
    'Front Panel Board',
    'Back Panel Board',
    'Power System Board',
    'Drive Backplane',
    'System Internal Expansion Board',
    'Other System Board',
    'Processor Board',
    'Power Unit',
    'Power Module',
    'Power Management',
    'Chassis Back Panel Board',
    'System Chassis',
    'Sub-Chassis',
    'Other Chassis Board',
    'Disk Drive Bay',
    'Peripheral Bay',
    'Device Bay',
    'Fan Device',
    'Cooling Unit',
    'Cable/Interconnect',
    'Memory Device',
    'System Management Software',
    'BIOS',
    'Operating System',
    'System Bus',
    'Group',
    'Remote Management Device',
    'External Environment',
    'Battery',
    'Processing Blade',
    'Connectivity Switch',
    'Processor/Memory Module',
    'I/O Module',
    'Processor/IO Module',
    'Management Controller Firmware',
    'IPMI Channel',
    'PCI Bus',
    'PCI Express Bus',
    'SCSI Bus (parallel)',
    'SATA/SAS Bus',
    'Processor/Front-Side Bus',
    'Real Time Clock(RTC)',
    'Reserved',
    'Air Inlet',
    'Reserved',
    'Reserved',
    'Reserved',
    'Reserved',
    'Reserved',
    'Reserved',
    'Reserved',
    'Reserved',
    'Air Inlet',
    'Processor',
    'Baseboard/Main System Board',
))

ENTITY_DEVICE_CODES = {
    0x01: 'Reserved',
    0x02: 'DS1624 temperature sensor',
    0x03: 'DS1621 temperature sensor',
    0x04: 'LM75 Temperature Sensor',
    0x05: 'Heceta ASIC',
    0x06: 'Reserved',
    0x07: 'Reserved',
    0x08: 'EEPROM, 24C01',
    0x09: 'EEPROM, 24C02',
    0x0a: 'EEPROM, 24C04',
    0x0b: 'EEPROM, 24C08',
    0x0c: 'EEPROM, 24C16',
    0x0d: 'EEPROM, 24C17',
    0x0e: 'EEPROM, 24C32',
    0x0f: 'EEPROM, 24C64',
    0x1000: 'IPMI FRU Inventory',
    0x1001: 'DIMM Memory ID',
    0x1002: 'IPMI FRU Inventory',
    0x1003: 'System Processor Cartridge FRU',
    0x11: 'Reserved',
    0x12: 'Reserved',
    0x13: 'Reserved',
    0x14: 'PCF 8570 256 byte RAM',
    0x15: 'PCF 8573 clock/calendar',
    0x16: 'PCF 8574A I/O Port',
    0x17: 'PCF 8583 clock/calendar',
    0x18: 'PCF 8593 clock/calendar',
    0x19: 'Clock calendar',
    0x1a: 'PCF 8591 A/D, D/A Converter',
    0x1b: 'I/O Port',
    0x1c: 'A/D Converter',
    0x1d: 'D/A Converter',
    0x1e: 'A/D, D/A Converter',
    0x1f: 'LCD Controller/Driver',
    0x20: 'Core Logic (Chip set) Device',
    0x21: 'LMC6874 Intelligent Battery controller',
    0x22: 'Intelligent Batter controller',
    0x23: 'Combo Management ASIC',
    0x24: 'Maxim 1617 Temperature Sensor',
    0xbf: 'Other/Unspecified',
}

GENERIC_EVENT_TYPES = TupleExt((
    TupleExt((
        # 0x00 - Nothing
    )),
    TupleExt((
        # 0x01 - Threshold Based States
        'Lower Non-critical going low',
        'Lower Non-critical going high',
        'Lower Critical going low',
        'Lower Critical going high',
        'Lower Non-recoverable going low',
        'Lower Non-recoverable going high',
        'Upper Non-critical going low',
        'Upper Non-critical going high',
        'Upper Critical going low',
        'Upper Critical going high',
        'Upper Non-recoverable going low',
        'Upper Non-recoverable going high',
    )),
    TupleExt((
        # 0x02 - DMI-based "usage state" States
        'Transition to Idle',
        'Transition to Active',
        'Transition to Busy',
    )),
    TupleExt((
        # 0x03 - DMI-based "usage state" States
        'State Deasserted',
        'State Asserted',
    )),
    TupleExt((
        # 0x04 - DMI-based "usage state" States
        'Predictive Failure Deasserted',
        'Predictive Failure Asserted',
    )),
    TupleExt((
        # 0x05 - DMI-based "usage state" States
        'Limit Not Exceeded',
        'Limit Exceeded',
    )),
    TupleExt((
        # 0x06 - Digital-Discrete Event States
        'Performance Met',
        'Performance Lags',
    )),
    TupleExt((
        # 0x07 - Severity Event States
        'Transition to OK',
        'Transition to Non-critical from OK',
        'Transition to Critical from less severe',
        'Transition to Non-recoverable from less severe',
        'Transition to Non-critical from more severe',
        'Transition to Critical from Non-recoverable',
        'Transition to Non-recoverable',
        'Monitor',
        'Informational',
    )),
    TupleExt((
        # 0x08 - Severity Event States
        'Device Absent',
        'Device Present',
    )),
    TupleExt((
        # 0x09 - Severity Event States
        'Device Disabled',
        'Device Enabled',
    )),
    TupleExt((
        # 0x0a - Availability Status States
        'Transition to Running',
        'Transition to In Test',
        'Transition to Power Off',
        'Transition to On Line',
        'Transition to Off Line',
        'Transition to Off Duty',
        'Transition to Degraded',
        'Transition to Power Save',
        'Install Error',
    )),
    TupleExt((
        # 0x0b - Redundancy States
        'Fully Redundant',
        'Redundancy Lost',
        'Redundancy Degraded',
        'Non-Redundant: Sufficient from Redundant',
        'Non-Redundant: Sufficient from Insufficient',
        'Non-Redundant: Insufficient Resources',
        'Redundancy Degraded from Fully Redundant',
        'Redundancy Degraded from Non-Redundant',
    )),
    TupleExt((
        # 0x0c - ACPI Device Power States
        'D0 Power State',
        'D1 Power State',
        'D2 Power State',
        'D3 Power State',
    )),
))

SPECIFIC_EVENT_TYPES = {
    0x05: TupleExt((
        #  Physical Security
        'General Chassis intrusion',
        'Drive Bay intrusion',
        'I/O Card area intrusion',
        'Processor area intrusion',
        'System unplugged from LAN',
        'Unauthorized dock',
        'FAN area intrusion',
    )),
    0x06: TupleExt((
        #  Platform Security
        'Front Panel Lockout violation attempted',
        'Pre-boot password violation - user password',
        'Pre-boot password violation - setup password',
        'Pre-boot password violation - network boot password',
        'Other pre-boot password violation',
        'Out-of-band access password violation',
    )),
    0x07: TupleExt((
        #  Processor
        'IERR',
        'Thermal Trip',
        'FRB1/BIST failure',
        'FRB2/Hang in POST failure',
        'FRB3/Processor startup/init failure',
        'Configuration Error',
        'SM BIOS Uncorrectable CPU-complex Error',
        'Presence detected',
        'Disabled',
        'Terminator presence detected',
        'Throttled',
        'Uncorrectable machine check exception',
        'Correctable machine check error',
    )),
    0x08: TupleExt((
        #  Power Supply
        'Presence detected',
        'Failure detected',
        'Predictive failure',
        'Power Supply AC lost',
        'AC lost or out-of-range',
        'AC out-of-range, but present',
        'Config Error: Vendor Mismatch',
        'Config Error: Revision Mismatch',
        'Config Error: Processor Missing',
        'Config Error: Power Supply Rating Mismatch',
        'Config Error: Voltage Rating Mismatch',
        'Config Error',
        'Power Supply Inactive',
    )),
    0x09: TupleExt((
        #  Power Unit
        'Power off/down',
        'Power cycle',
        '240VA power down',
        'Interlock power down',
        'AC lost',
        'Soft-power control failure',
        'Failure detected',
        'Predictive failure',
    )),
    0x0c: TupleExt((
        #  Memory
        'Correctable ECC',
        'Uncorrectable ECC',
        'Parity',
        'Memory Scrub Failed',
        'Memory Device Disabled',
        'Correctable ECC logging limit reached',
        'Presence Detected',
        'Configuration Error',
        'Spare',
        'Throttled',
        'Critical Overtemperature',
    )),
    0x0d: TupleExt((
        #  Drive Slot
        'Drive Present',
        'Drive Fault',
        'Predictive Failure',
        'Hot Spare',
        'Parity Check In Progress',
        'In Critical Array',
        'In Failed Array',
        'Rebuild In Progress',
        'Rebuild Aborted',
    )),
    0x0f: TupleExt((
        #  System Firmware Progress
        'Unspecified',
        'No system memory installed',
        'No usable system memory',
        'Unrecoverable IDE device failure',
        'Unrecoverable system-board failure',
        'Unrecoverable diskette failure',
        'Unrecoverable hard-disk controller failure',
        'Unrecoverable PS/2 or USB keyboard failure',
        'Removable boot media not found',
        'Unrecoverable video controller failure',
        'No video device selected',
        'BIOS corruption detected',
        'CPU voltage mismatch',
        'CPU speed mismatch failure',
        'Unknown Error',
        'Unspecified',
        'Memory initialization',
        'Hard-disk initialization',
        'Secondary CPU Initialization',
        'User authentication',
        'User-initiated system setup',
        'USB resource configuration',
        'PCI resource configuration',
        'Option ROM initialization',
        'Video initialization',
        'Cache initialization',
        'SMBus initialization',
        'Keyboard controller initialization',
        'Management controller initialization',
        'Docking station attachment',
        'Enabling docking station',
        'Docking station ejection',
        'Disabling docking station',
        'Calling operating system wake-up vector',
        'System boot initiated',
        'Motherboard initialization',
        'reserved',
        'Floppy initialization',
        'Keyboard test',
        'Pointing device test',
        'Primary CPU initialization',
        'Unknown Hang',
        'Unspecified',
        'Memory initialization',
        'Hard-disk initialization',
        'Secondary CPU Initialization',
        'User authentication',
        'User-initiated system setup',
        'USB resource configuration',
        'PCI resource configuration',
        'Option ROM initialization',
        'Video initialization',
        'Cache initialization',
        'SMBus initialization',
        'Keyboard controller initialization',
        'Management controller initialization',
        'Docking station attachment',
        'Enabling docking station',
        'Docking station ejection',
        'Disabling docking station',
        'Calling operating system wake-up vector',
        'System boot initiated',
        'Motherboard initialization',
        'reserved',
        'Floppy initialization',
        'Keyboard test',
        'Pointing device test',
        'Primary CPU initialization',
        'Unknown Progress',
    )),
    0x10: TupleExt((
        #  Event Logging Disabled
        'Correctable memory error logging disabled',
        'Event logging disabled',
        'Log area reset/cleared',
        'All event logging disabled',
        'Log full',
        'Log almost full',
    )),
    0x11: TupleExt((
        #  Watchdog 1
        'BIOS Reset',
        'OS Reset',
        'OS Shut Down',
        'OS Power Down',
        'OS Power Cycle',
        'OS NMI/Diag Interrupt',
        'OS Expired',
        'OS pre-timeout Interrupt',
    )),
    0x12: TupleExt((
        #  System Event
        'System Reconfigured',
        'OEM System boot event',
        'Undetermined system hardware failure',
        'Entry added to auxiliary log',
        'PEF Action',
        'Timestamp Clock Sync',
    )),
    0x13: TupleExt((
        #  Critical Interrupt
        'NMI/Diag Interrupt',
        'Bus Timeout',
        'I/O Channel check NMI',
        'Software NMI',
        'PCI PERR',
        'PCI SERR',
        'EISA failsafe timeout',
        'Bus Correctable error',
        'Bus Uncorrectable error',
        'Fatal NMI',
        'Bus Fatal Error',
        'Bus Degraded',
    )),
    0x14: TupleExt((
        #  Button
        'Power Button pressed',
        'Sleep Button pressed',
        'Reset Button pressed',
        'FRU Latch',
        'FRU Service',
    )),
    0x19: TupleExt((
        #  Chip Set
        'Soft Power Control Failure',
        'Thermal Trip',
    )),
    0x1b: TupleExt((
        #  Cable/Interconnect
        'Connected',
        'Config Error',
    )),
    0x1d: TupleExt((
        #  System Boot Initiated
        'Initiated by power up',
        'Initiated by hard reset',
        'Initiated by warm reset',
        'User requested PXE boot',
        'Automatic boot to diagnostic',
        'OS initiated hard reset',
        'OS initiated warm reset',
        'System Restart',
    )),
    0x1e: TupleExt((
        #  Boot Error
        'No bootable media',
        'Non-bootable disk in drive',
        'PXE server not found',
        'Invalid boot sector',
        'Timeout waiting for selection',
    )),
    0x1f: TupleExt((
        #  OS Boot
        'A: boot completed',
        'C: boot completed',
        'PXE boot completed',
        'Diagnostic boot completed',
        'CD-ROM boot completed',
        'ROM boot completed',
        'boot completed - device not specified',
        'Installation started',
        'Installation completed',
        'Installation aborted',
        'Installation failed',
    )),
    0x20: TupleExt((
        #  OS Stop/Shutdown
        'Error during system startup',
        'Run-time critical stop',
        'OS graceful stop',
        'OS graceful shutdown',
        'PEF initiated soft shutdown',
        'Agent not responding',
    )),
    0x21: TupleExt((
        #  Slot/Connector
        'Fault Status',
        'Identify Status',
        'Device Installed',
        'Ready for Device Installation',
        'Ready for Device Removal',
        'Slot Power is Off',
        'Device Removal Request',
        'Interlock',
        'Slot is Disabled',
        'Spare Device',
    )),
    0x22: TupleExt((
        #  System ACPI Power State
        'S0/G0: working',
        'S1: sleeping with system hw & processor context maintained',
        'S2: sleeping, processor context lost',
        'S3: sleeping, processor & hw context l',
        'S4: non-volatile sleep/suspend-to-disk',
        'S5/G2: soft-off',
        'S4/S5: soft-off',
        'G3: mechanical off',
        'Sleeping in S1/S2/S3 state',
        'G1: sleeping',
        'S5: entered by override',
        'Legacy ON state',
        'Legacy OFF state',
        'Unknown',
    )),
    0x23: TupleExt((
        #  Watchdog 2
        'Timer expired',
        'Hard reset',
        'Power down',
        'Power cycle',
        'reserved',
        'reserved',
        'reserved',
        'reserved',
        'Timer interrupt',
    )),
    0x24: TupleExt((
        #  Platform Alert
        'Platform generated page',
        'Platform generated LAN alert',
        'Platform Event Trap generated',
        'Platform generated SNMP trap, OEM format',
    )),
    0x25: TupleExt((
        #  Entity Presence
        'Present',
        'Absent',
        'Disabled',
    )),
    0x27: TupleExt((
        #  LAN
        'Heartbeat Lost',
        'Heartbeat',
    )),
    0x28: TupleExt((
        #  Management Subsystem Health
        'Sensor access degraded or unavailable',
        'Controller access degraded or unavailable',
        'Management controller off-line',
        'Management controller unavailable',
        'Sensor failure',
        'FRU failure',
    )),
    0x29: TupleExt((
        #  Battery
        'Low',
        'Failed',
        'Presence Detected',
    )),
    0x2b: TupleExt((
        #  Version Change
        'Hardware change detected',
        'Firmware or software change detected',
        'Firmware or software change detected, Mngmt Ctrl Dev Id',
        'Firmware or software change detected, Mngmt Ctrl Firm Rev',
        'Firmware or software change detected, Mngmt Ctrl Dev Rev',
        'Firmware or software change detected, Mngmt Ctrl Manuf Id',
        'Firmware or software change detected, Mngmt Ctrl IPMI Vers',
        'Firmware or software change detected, Mngmt Ctrl Aux Firm Id',
        'Firmware or software change detected, Mngmt Ctrl Firm Boot Block',
        'Firmware or software change detected, Mngmt Ctrl Other',
        'Firmware or software change detected, BIOS/EFI change',
        'Firmware or software change detected, SMBIOS change',
        'Firmware or software change detected, O/S change',
        'Firmware or software change detected, O/S loader change',
        'Firmware or software change detected, Service Diag change',
        'Firmware or software change detected, Mngmt SW agent change',
        'Firmware or software change detected, Mngmt SW App change',
        'Firmware or software change detected, Mngmt SW Middle',
        'Firmware or software change detected, Prog HW Change (FPGA)',
        'Firmware or software change detected, board/FRU module change',
        'Firmware or software change detected, board/FRU component change',
        'Firmware or software change detected, board/FRU replace equ ver',
        'Firmware or software change detected, board/FRU replace new ver',
        'Firmware or software change detected, board/FRU replace old ver',
        'Firmware or software change detected, board/FRU HW conf change',
        'Hardware incompatibility detected',
        'Firmware or software incompatibility detected',
        'Invalid or unsupported hardware version',
        'Invalid or unsupported firmware or software version',
        'Hardware change success',
        'Firmware or software change success',
        'Firmware or software change success, Mngmt Ctrl Dev Id',
        'Firmware or software change success, Mngmt Ctrl Firm Rev',
        'Firmware or software change success, Mngmt Ctrl Dev Rev',
        'Firmware or software change success, Mngmt Ctrl Manuf Id',
        'Firmware or software change success, Mngmt Ctrl IPMI Vers',
        'Firmware or software change success, Mngmt Ctrl Aux Firm Id',
        'Firmware or software change success, Mngmt Ctrl Firm Boot Block',
        'Firmware or software change success, Mngmt Ctrl Other',
        'Firmware or software change success, BIOS/EFI change',
        'Firmware or software change success, SMBIOS change',
        'Firmware or software change success, O/S change',
        'Firmware or software change success, O/S loader change',
        'Firmware or software change success, Service Diag change',
        'Firmware or software change success, Mngmt SW agent change',
        'Firmware or software change success, Mngmt SW App change',
        'Firmware or software change success, Mngmt SW Middle',
        'Firmware or software change success, Prog HW Change (FPGA)',
        'Firmware or software change success, board/FRU module change',
        'Firmware or software change success, board/FRU component change',
        'Firmware or software change success, board/FRU replace equ ver',
        'Firmware or software change success, board/FRU replace new ver',
        'Firmware or software change success, board/FRU replace old ver',
        'Firmware or software change success, board/FRU HW conf change',
    )),
    0x2c: TupleExt((
        #  FRU State
        'Not Installed',
        'Inactive',
        'Activation Requested',
        'Activation in Progress',
        'Active',
        'Deactivation Requested',
        'Deactivation in Progress',
        'Communication lost',
    )),
    0xF0: TupleExt((
        #  PICMG FRU Hot Swap
        'Transition to M0',
        'Transition to M1',
        'Transition to M2',
        'Transition to M3',
        'Transition to M4',
        'Transition to M5',
        'Transition to M6',
        'Transition to M7',
    )),
    0xF1: TupleExt((
        #  PICMG IPMB Physical Link
        'IPMB-A disabled, IPMB-B disabled',
        'IPMB-A enabled, IPMB-B disabled',
        'IPMB-A disabled, IPMB-B enabled',
        'IPMB-A enabled, IPMB-B enabled',
    )),
    0xF2: TupleExt((
        #  PICMG IPMB Physical Link
        'Module Handle Closed',
        'Module Handle Opened',
        'Quiesced',
    )),
}

USER_PRIVS = {0: 'reserved', 1: 'Callback', 2: 'User', 3: 'Operator', 4: 'Administrator',
              5: 'OEM', 0x0f: 'No Access'}

AUTH_TYPES = {'none': 1, 'md2': 2, 'md5': 4, 'password': 16, 'oem': 32}

CHASSIS_TYPES = TupleExt((
    'Unknown', 'Other', 'Unknown', 'Desktop', 'Low Profile Desktop', 'Pizza Box',
    'Mini Tower', 'Tower', 'Portable', 'Laptop', 'Notebook', 'Hand Held', 
    'Docking Station', 'All in One', 'Sub Notebook', 'Space-saving', 'Lunch Box',
    'Main Server Chassis', 'Expansion Chassis', 'SubChassis', 'Bus Expansion Chassis',
    'Peripheral Chassis', 'RAID Chassis', 'Rack Mount Chassis', 'Sealed-case PC', 
    'Multi-system chassis', 'Compact PCI', 'Advanced PCI', 'Blade', 'Blade Enclosure',
    'Tablet', 'Convertible', 'Detachable', 'IoT Gateway', 'Embedded PC', 'Mini PC',
    'Stick PC',
))




#
# Copyright (c) 2020, Hyve Design Solutions Corporation.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are 
# met:
#
# 1. Redistributions of source code must retain the above copyright 
#    notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright 
#    notice, this list of conditions and the following disclaimer in the 
#    documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of Hyve Design Solutions Corporation nor the names 
#    of its contributors may be used to endorse or promote products 
#    derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY HYVE DESIGN SOLUTIONS CORPORATION AND 
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, 
# BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND 
# FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 
# HYVE DESIGN SOLUTIONS CORPORATION OR CONTRIBUTORS BE LIABLE FOR ANY 
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL 
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS 
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
# STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING 
# IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
# POSSIBILITY OF SUCH DAMAGE.
#
import sys
from .. util import PyTest
from ._raw import do_raw
from ._lan import do_lan
from ._mc import do_mc
from ._chassis import do_chassis
from ._power import do_power
from ._sdr import do_sdr
from ._sel import do_sel
from ._sensor import do_sensor
from ._user import do_user
from ._fru import do_fru

__all__ = [
    'PyCmds',
]

PY_CMDS = {
    'mc': do_mc,
    'chassis': do_chassis,
    'power': do_power,
    'sdr': do_sdr,
    'sel': do_sel,
    'sensor': do_sensor,
    'user': do_user,
    'lan': do_lan,
    'fru': do_fru,
    'raw': do_raw,
}

class StrEx(str):
    def __init__(self):
        self.print_str = ''

    def __str__(self):
        return self.print_str

    def write(self, text):        
        self.print_str += text

    def get_str(self):
        return self.print_str

    def reset(self):
        self.print_str = '' 

class PyCmds(PyTest):
    def __init__(self, opts_overwrite=None, print_file=sys.stdout, keep_alive=False):
        super(PyCmds, self).__init__(opts_overwrite, keep_alive=keep_alive)
        self.print_file = print_file

    @staticmethod
    def print_usage(prg=None):
        if prg is not None:
            print('Usage: {0} <command> [<options>]'.format(prg))
        print('Commands:')
        print('    ping')
        for i in PY_CMDS.keys():
            print('    {0}'.format(i))
        print('    help')

    from ._common import print
    from ._common import print_rsp

    def exec_command(self, cmd):
        if type(cmd) is str:
            args = cmd.split(' ')
            args = [x.strip() for x in args]
        else:
            args = cmd
        
        command = args[0]
        args = args[1:]

        if command == 'help':
            PyCmds.print_usage()
        elif command == 'ping':
            try:
                self.intf.ping()
                print('Success')
            except BaseException as e:
                print(e)
        elif command not in PY_CMDS.keys():
            print('Invalid command:', command)
        else:
            PY_CMDS[command](self, args)



#
# Copyright (c) 2020, Hyve Design Solutions Corporation.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are 
# met:
#
# 1. Redistributions of source code must retain the above copyright 
#    notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright 
#    notice, this list of conditions and the following disclaimer in the 
#    documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of Hyve Design Solutions Corporation nor the names 
#    of its contributors may be used to endorse or promote products 
#    derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY HYVE DESIGN SOLUTIONS CORPORATION AND 
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, 
# BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND 
# FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 
# HYVE DESIGN SOLUTIONS CORPORATION OR CONTRIBUTORS BE LIABLE FOR ANY 
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL 
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS 
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
# STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING 
# IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
# POSSIBILITY OF SUCH DAMAGE.
#
import os, sys
from optparse import OptionParser
from . config import PyConfig
from .. import intf

__all__ = [
    'checksum',
    'PyTest',
    'config',
    'exception',
]

def checksum(data, fmt=bytes):
    chksum = 0
    for i in data:
        chksum += i
    chksum &= 0xff
    if chksum != 0: chksum = 256 - chksum

    if fmt == bytes:
        return chksum.to_bytes(1, byteorder='little')
    else:
        return chksum

class PyTest:
    def __init__(self, opts_overwrite=None, ping_only=False, keep_alive=False,
                     raise_except=False):
        self.ping_only = ping_only

        try:
            # read the config file
            config_file = os.path.join(os.getenv('HOME'), '.config', 'pyipmi', 'pyipmi.conf')
            self.conf = PyConfig()
            opts, w_flag = self.conf.parse_config(config_file)  

            # command line option to overwrite the settings in the config file
            if opts_overwrite:
                opts = self.conf.overwrite_config(opts, opts_overwrite)
                if w_flag or opts_overwrite.force:  
                    self.conf.write_config(config_file, opts)

            # create the interface object
            self.intf = intf.init(opts, ping_only, keep_alive)

        except BaseException as e:
            if raise_except:
                raise BaseException(e)
            else:
                print(e)
                exit(1)

    def print_rsp(self, rsp):
        print(' '.join(('{0:02x}'.format(i) for i in rsp)))

    def run_commands(self, argv=None):
        if self.ping_only:
            # RMCP Ping
            self.intf.ping()
            print('Success')
        else:
            # IPMI command: Get Device ID (NetFn=App, CMD=1)
            req = [6, 1]
            rsp = self.intf.issue_raw_cmd(req)
            self.print_rsp(rsp)

    def run(self, argv=None):
        ret = 1
        try:
            self.run_commands(argv)
            ret = 0
        except BaseException as e:
            print(e)
        finally:
            return ret



#
# Copyright (c) 2020, Hyve Design Solutions Corporation.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are 
# met:
#
# 1. Redistributions of source code must retain the above copyright 
#    notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright 
#    notice, this list of conditions and the following disclaimer in the 
#    documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of Hyve Design Solutions Corporation nor the names 
#    of its contributors may be used to endorse or promote products 
#    derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY HYVE DESIGN SOLUTIONS CORPORATION AND 
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, 
# BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND 
# FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 
# HYVE DESIGN SOLUTIONS CORPORATION OR CONTRIBUTORS BE LIABLE FOR ANY 
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL 
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS 
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
# STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING 
# IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
# POSSIBILITY OF SUCH DAMAGE.
#
import struct
from . import IPMI_Message
from .. util.exception import PyMesgExcept, PyMesgCCExcept

NETFN_STORAGE = 0x0a

class GetFruAreaInfo(IPMI_Message):
    def __init__(self, fru_id):
        super(GetFruAreaInfo, self).__init__(NETFN_STORAGE, 0x10)
        self.req_data = struct.pack('B', fru_id)

    def unpack(self, rsp):
        return super(GetFruAreaInfo, self).unpack(rsp, '<HB')

class ReadFru(IPMI_Message):
    def __init__(self, fru_id, offset, count):
        super(ReadFru, self).__init__(NETFN_STORAGE, 0x11)
        offset_ls = offset & 0xff
        offset_ms = (offset >> 8) & 0xff
        self.req_data = struct.pack('BBBB', fru_id, offset_ls, offset_ms, count)

    def unpack(self, rsp):
        return super(ReadFru, self).unpack(rsp)

class WriteFru(IPMI_Message):
    def __init__(self, fru_id, offset, data):
        super(WriteFru, self).__init__(NETFN_STORAGE, 0x12)
        offset_ls = offset & 0xff
        offset_ms = (offset >> 8) & 0xff
        self.req_data = struct.pack('BBB', fru_id, offset_ls, offset_ms) + data

    def unpack(self, rsp):
        try:
            return super(WriteFru, self).unpack(rsp, 'B')            
        except PyMesgCCExcept as e:
            if e.cc == 0x80:
                print('Failed to write FRU.  Write-protected offset (CC=80h).')
            elif e.cc == 0x81:
                print('Failed to write FRU.  FRU device busy (CC=81h).')
            else:
                print(e)

class GetSDRRepoInfo(IPMI_Message):
    def __init__(self):
        super(GetSDRRepoInfo, self).__init__(NETFN_STORAGE, 0x20)
        self.req_data = None

    def unpack(self, rsp):
        return super(GetSDRRepoInfo, self).unpack(rsp, '<BHHLLB')

class GetSDRRepoAllocInfo(IPMI_Message):
    def __init__(self):
        super(GetSDRRepoAllocInfo, self).__init__(NETFN_STORAGE, 0x21)
        self.req_data = None

    def unpack(self, rsp):
        return super(GetSDRRepoAllocInfo, self).unpack(rsp, '<HHHHB')

class RevSDR(IPMI_Message):
    def __init__(self):
        super(RevSDR, self).__init__(NETFN_STORAGE, 0x22)
        self.req_data = None

    def unpack(self, rsp):
        return super(RevSDR, self).unpack(rsp, '<2s')

class GetSDR(IPMI_Message):
    def __init__(self, rev_id, rec_id, offset, bytes_to_read):
        super(GetSDR, self).__init__(NETFN_STORAGE, 0x23)
        self.req_data = struct.pack('<2s2sBB', rev_id, rec_id, offset, bytes_to_read)

    def unpack(self, rsp):
        return super(GetSDR, self).unpack(rsp)

class AddSDR(IPMI_Message):
    def __init__(self, data):
        super(AddSDR, self).__init__(NETFN_STORAGE, 0x24)
        self.req_data = data

    def unpack(self, rsp):
        return super(AddSDR, self).unpack(rsp, '<2s')

class PartialAddSDR(IPMI_Message):
    def __init__(self, rev_id, rec_id, offset, stat, data):
        super(PartialAddSDR, self).__init__(NETFN_STORAGE, 0x25)
        self.req_data = struct.pack('<2s2sBB', rev_id, rec_id, offset, stat) + data

    def unpack(self, rsp):
        return super(PartialAddSDR, self).unpack(rsp, '<2s')

class ClearSDR(IPMI_Message):
    def __init__(self, rev, stat):
        super(ClearSDR, self).__init__(NETFN_STORAGE, 0x27)
        self.req_data = struct.pack('2sBBBB', rev, 0x43, 0x4c, 0x52, stat)

    def unpack(self, rsp):
        return super(ClearSDR, self).unpack(rsp, 'B')

class GetSELInfo(IPMI_Message):
    def __init__(self):
        super(GetSELInfo, self).__init__(NETFN_STORAGE, 0x40)
        self.req_data = None

    def unpack(self, rsp):
        return super(GetSELInfo, self).unpack(rsp, '<BHHLLB')

class GetSELAllocInfo(IPMI_Message):
    def __init__(self):
        super(GetSELAllocInfo, self).__init__(NETFN_STORAGE, 0x41)
        self.req_data = None

    def unpack(self, rsp):
        return super(GetSELAllocInfo, self).unpack(rsp, '<HHHHB')

class RevSEL(IPMI_Message):
    def __init__(self):
        super(RevSEL, self).__init__(NETFN_STORAGE, 0x42)
        self.req_data = None

    def unpack(self, rsp):
        return super(RevSEL, self).unpack(rsp, '<2s')

class GetSELEntry(IPMI_Message):
    def __init__(self, rec_id):
        super(GetSELEntry, self).__init__(NETFN_STORAGE, 0x43)
        self.req_data = struct.pack('<2s2sBB', b'\0', rec_id, 0, 0xff)

    def unpack(self, rsp):
        return super(GetSELEntry, self).unpack(rsp, '<2sHBLHBBBBBBB')

class ClearSEL(IPMI_Message):
    def __init__(self, rev, stat):
        super(ClearSEL, self).__init__(NETFN_STORAGE, 0x47)
        self.req_data = struct.pack('2sBBBB', rev, 0x43, 0x4c, 0x52, stat)

    def unpack(self, rsp):
        return super(ClearSEL, self).unpack(rsp, 'B')

class GetSELTime(IPMI_Message):
    def __init__(self):
        super(GetSELTime, self).__init__(NETFN_STORAGE, 0x48)
        self.req_data = None

    def unpack(self, rsp):
        return super(GetSELTime, self).unpack(rsp, '<L')




#
# Copyright (c) 2020, Hyve Design Solutions Corporation.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are 
# met:
#
# 1. Redistributions of source code must retain the above copyright 
#    notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright 
#    notice, this list of conditions and the following disclaimer in the 
#    documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of Hyve Design Solutions Corporation nor the names 
#    of its contributors may be used to endorse or promote products 
#    derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY HYVE DESIGN SOLUTIONS CORPORATION AND 
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, 
# BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND 
# FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 
# HYVE DESIGN SOLUTIONS CORPORATION OR CONTRIBUTORS BE LIABLE FOR ANY 
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL 
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS 
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
# STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING 
# IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
# POSSIBILITY OF SUCH DAMAGE.
#
from .. mesg.ipmi_chassis import GetChsStat, ChsCtrl
from . _common import do_command

def _power_status(self, argv):
    t1 = self.intf.issue_cmd(GetChsStat)
    pstat = 'on' if t1[0] & 1 else 'off'
    self.print('Chassis Power is', pstat)

def _power_on(self, argv):
    self.intf.issue_cmd(ChsCtrl, 1)
    self.print('Chassis Power Control: Up/On')

def _power_off(self, argv):
    self.intf.issue_cmd(ChsCtrl, 0)
    self.print('Chassis Power Control: Down/Off')

def _power_cycle(self, argv):
    self.intf.issue_cmd(ChsCtrl, 2)
    self.print('Chassis Power Control: Cycle')

def _power_reset(self, argv):
    self.intf.issue_cmd(ChsCtrl, 3)
    self.print('Chassis Power Control: Hard Reset')

def _power_diag(self, argv):
    self.intf.issue_cmd(ChsCtrl, 4)
    self.print('Chassis Power Control: Pulse Diagnostic Interrupt')

def _power_soft(self, argv):
    self.intf.issue_cmd(ChsCtrl, 5)
    self.print('Chassis Power Control: Soft Shutdown')

def help_power(self, argv=None, context=0):
    self.print('Power Commands:')
    for cmd in POWER_CMDS.keys():
        self.print('    {0}'.format(cmd))

POWER_CMDS = {
    'status': _power_status,
    'on': _power_on,
    'off': _power_off,
    'cycle': _power_cycle,
    'reset': _power_reset,
    'diag': _power_diag,
    'soft': _power_soft,
    'help': help_power,
}

def do_power(self, argv):
    do_command(self, argv, help_power, POWER_CMDS)



#
# Copyright (c) 2020, Hyve Design Solutions Corporation.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are 
# met:
#
# 1. Redistributions of source code must retain the above copyright 
#    notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright 
#    notice, this list of conditions and the following disclaimer in the 
#    documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of Hyve Design Solutions Corporation nor the names 
#    of its contributors may be used to endorse or promote products 
#    derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY HYVE DESIGN SOLUTIONS CORPORATION AND 
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, 
# BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND 
# FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 
# HYVE DESIGN SOLUTIONS CORPORATION OR CONTRIBUTORS BE LIABLE FOR ANY 
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL 
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS 
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
# STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING 
# IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
# POSSIBILITY OF SUCH DAMAGE.
#
import os, struct
import Crypto.Cipher.AES, Crypto.Hash.HMAC, Crypto.Hash.MD2, Crypto.Hash.MD5
import Crypto.Hash.SHA, Crypto.Hash.SHA256

# Common
def conv_str2bytes(org):
    if type(org) is bytes: return org
    elif type(org) is not str: return None
    if org == '': return None
    return bytes(org, 'latin_1')

# RMCP
AUTH_NONE = 0
AUTH_MD2 = 1
AUTH_MD5 = 2
AUTH_PASSWD = 4
AUTH_OEM = 5
AUTH_RMCPP = 6

RMCP_AUTHS = {
    'none': AUTH_NONE,
    'md2': AUTH_MD2,
    'md5': AUTH_MD5,
    'password': AUTH_PASSWD,
    'oem': AUTH_OEM,
}

def cal_auth_code_15(auth, pwd, sseq=0, sid='\0', data=None):
    auth_code = None
    if pwd is not None:
        pwd_b = struct.pack('16s', pwd)
    else:
        pwd_b = struct.pack('16s', b'\0')
    H = None

    if auth == AUTH_PASSWD:
        return pwd_b
    elif auth == AUTH_MD2:
        H = Crypto.Hash.MD2
    elif auth == AUTH_MD5:
        H = Crypto.Hash.MD5
    else:
        return None

    sid_b = struct.pack('4s', sid)
    sseq_b = struct.pack('<L', sseq)

    data_in = pwd_b + sid_b + data + sseq_b + pwd_b
    h1 = H.new(data_in)
    auth_code = h1.digest()

    return auth_code

# RMCP+
# Authentication
RAKP_NONE = 0
RAKP_HMAC_SHA1 = 1
RAKP_HMAC_MD5 = 2
RAKP_HMAC_SHA256 = 3

# Integrity 
HMAC_SHA1_96 = 1
HMAC_MD5_128 = 2
MD5_128 = 3
HMAC_SHA256_128 = 4

# Confidentiality 
AES_CBC_128 = 1
#XRC4_128 = 2   # not support
#XRC4_40 = 3    # not support

# RMCP+ Cipher Suites
RMCPP_CIPHERS = {
    # ID: (Auth, Inte, Conf)
    0: (RAKP_NONE, RAKP_NONE, RAKP_NONE),
    1: (RAKP_HMAC_SHA1, RAKP_NONE, RAKP_NONE),
    2: (RAKP_HMAC_SHA1, HMAC_SHA1_96, RAKP_NONE),
    3: (RAKP_HMAC_SHA1, HMAC_SHA1_96, AES_CBC_128),
    #4: (RAKP_HMAC_SHA1, HMAC_SHA1_96, XRC4_128),
    #5: (RAKP_HMAC_SHA1, HMAC_SHA1_96, XRC4_40),
    6: (RAKP_HMAC_MD5, RAKP_NONE, RAKP_NONE),
    7: (RAKP_HMAC_MD5, HMAC_MD5_128, RAKP_NONE),
    8: (RAKP_HMAC_MD5, HMAC_MD5_128, AES_CBC_128),
    #9: (RAKP_HMAC_MD5, HMAC_MD5_128, XRC4_128),
    #10: (RAKP_HMAC_MD5, HMAC_MD5_128, XRC4_40),
    11: (RAKP_HMAC_MD5, MD5_128, RAKP_NONE),
    12: (RAKP_HMAC_MD5, MD5_128, AES_CBC_128),
    #13: (RAKP_HMAC_MD5, MD5_128, XRC4_128),
    #14: (RAKP_HMAC_MD5, MD5_128, XRC4_40),
    15: (RAKP_HMAC_SHA256, RAKP_NONE, RAKP_NONE),
    16: (RAKP_HMAC_SHA256, HMAC_SHA256_128, RAKP_NONE),
    17: (RAKP_HMAC_SHA256, HMAC_SHA256_128, AES_CBC_128),
    #18: (RAKP_HMAC_SHA256, HMAC_SHA256_128, XRC4_128),
    #19: (RAKP_HMAC_SHA256, HMAC_SHA256_128, XRC4_40),
}

def get_cipher_tuple(cipher):
    if cipher not in RMCPP_CIPHERS.keys():
        return None
    return RMCPP_CIPHERS[cipher]

def get_cipher(cipher_tuple):
    ret = None
    for c, t in zip(RMCPP_CIPHERS.keys(), RMCPP_CIPHERS.values()):
        if t == cipher_tuple:
            ret = c
            break
    return ret

def cal_auth_code(algo, key, data):
    if algo == RAKP_HMAC_SHA1:
        H = Crypto.Hash.HMAC.new(key, data, Crypto.Hash.SHA)
    elif algo == RAKP_HMAC_MD5:
        H = Crypto.Hash.HMAC.new(key, data, Crypto.Hash.MD5)
    elif algo == RAKP_HMAC_SHA256:
        H = Crypto.Hash.HMAC.new(key, data, Crypto.Hash.SHA256)
    else:
        return None
        
    return H.digest()

def cal_inte_check(algo, key, data):
    trunc = None

    if algo == HMAC_SHA1_96:
        H = Crypto.Hash.HMAC.new(key, data, Crypto.Hash.SHA)
        trunc = 12
    elif algo == HMAC_MD5_128:
        H = Crypto.Hash.HMAC.new(key, data, Crypto.Hash.MD5)
    elif algo == HMAC_SHA256_128:
        H = Crypto.Hash.HMAC.new(key, data, Crypto.Hash.SHA256)
        trunc = 16
    elif algo == MD5_128:
        H = Crypto.Hash.MD5.new(key + data + key)
    else:
        return None
        
    ret = H.digest()
    if trunc is not None: ret = ret[:trunc]

    return ret 
    
def get_inte_len(algo):
    ret = 0

    if algo == HMAC_SHA1_96:
        ret = 12
    elif algo == HMAC_MD5_128:
        ret = 16
    elif algo == HMAC_SHA256_128:
        ret = 16
    elif algo == MD5_128:
        ret = 16

    return ret

def map_auth2inte(auth):
    inte = None

    if auth == RAKP_HMAC_SHA1:
        inte = HMAC_SHA1_96
    elif auth == RAKP_HMAC_MD5:
        inte = HMAC_MD5_128
    elif auth == RAKP_HMAC_SHA256:
        inte = HMAC_SHA256_128

    return inte

def encrypt_aes_128_cbc(key, data):
    iv = os.urandom(16)
    rem = (len(data) + 1) & 0x0f
    if rem != 0:
        # need padding bytes        
        pad_len = 16 - rem
        data += bytes((x for x in range(1, pad_len + 1)))
    else:
        pad_len = 0

    data += struct.pack('B', pad_len)
    E = Crypto.Cipher.AES.new(key[:16], Crypto.Cipher.AES.MODE_CBC, iv)
    
    return iv + E.encrypt(data)

def decrypt_aes_128_cbc(key, data):
    iv = data[:16]
    D = Crypto.Cipher.AES.new(key[:16], Crypto.Cipher.AES.MODE_CBC, iv)
    data = D.decrypt(data[16:])
    pad_len = data[-1]

    return data[:-(pad_len+1)]



#
# Copyright (c) 2020, Hyve Design Solutions Corporation.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are 
# met:
#
# 1. Redistributions of source code must retain the above copyright 
#    notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright 
#    notice, this list of conditions and the following disclaimer in the 
#    documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of Hyve Design Solutions Corporation nor the names 
#    of its contributors may be used to endorse or promote products 
#    derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY HYVE DESIGN SOLUTIONS CORPORATION AND 
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, 
# BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND 
# FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 
# HYVE DESIGN SOLUTIONS CORPORATION OR CONTRIBUTORS BE LIABLE FOR ANY 
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL 
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS 
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
# STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING 
# IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
# POSSIBILITY OF SUCH DAMAGE.
#
import os, configparser
from optparse import OptionParser
from . exception import PyConfExcept

class PyConfig:
    valid_opts = {
        'global': {
            'interface': (('lan', 'lanplus', 'kcs'), 'lanplus'),
            'host': (str, 'localhost'),
            'port': (int, 623),
        },

        'lan': {
            'user': (str, 'root'),
            'password': (str,  'root123'),
            'auth': (('none', 'md2', 'md5', 'password', 'oem'), 'md5'),
            'no_ping': (bool, True),
        },

        'lanplus': {
            'user': (str, 'admin'),
            'password': (str,  'admin123'),
            'cipher_suite': (int, 3),
            'priv': (int, 4),
            'kg': (str, ''),
            'no_ping': (bool, True),
        },

        'kcs': {
            'bmc_addr': (int, 0x20),
            'dev_num': (int, 0),
        },
    }

    def __init__(self):
        self.config = configparser.ConfigParser()

    def write_config(self, config_file, opts=None):
        if opts:
            for key in opts.keys():
                self.config[key] = opts[key]

        with open(config_file, 'w') as configfile:
            configfile.write('###### Generic Configuration File for PyIPMI ######\n')
            self.config.write(configfile)
            configfile.flush()
            configfile.close()

    def write_default_config(self, config_file):
        # check if the file path exists
        path = os.path.dirname(config_file)
        if not os.path.exists(path):
            os.mkdir(path, 0o755)

        # load default values
        for k1 in PyConfig.valid_opts.keys():
            d1 = {}
            for k2 in PyConfig.valid_opts[k1].keys():
                if PyConfig.valid_opts[k1][k2][0] is bytes:
                    d1[k2] = PyConfig.valid_opts[k1][k2][1].decode()
                else:
                    d1[k2] = str(PyConfig.valid_opts[k1][k2][1])
            
            self.config[k1] = d1

        # write to the config file
        self.write_config(config_file)

    def parse_config(self, config_file):
        # assign default values
        d1 = {}
        for k1 in PyConfig.valid_opts.keys():
            d2 = {}
            for k2 in PyConfig.valid_opts[k1].keys():
                d2[k2] = PyConfig.valid_opts[k1][k2][1]
            d1[k1] = d2

        flag = False

        # check if the config file is valid
        if not os.path.isfile(config_file):
            # config file not exist, create it
            self.write_default_config(config_file)
            flag = True
        else:
            # read the config file
            self.config.read(config_file)

            # assign values from the config file
            for s in self.config.sections():
                if not s in PyConfig.valid_opts.keys(): continue
                for key in self.config[s]:
                    if not key in PyConfig.valid_opts[s].keys(): continue

                    if isinstance(PyConfig.valid_opts[s][key][0], tuple):
                        if self.config[s][key] in PyConfig.valid_opts[s][key][0]:
                            d1[s][key] = self.config[s][key]
                    elif PyConfig.valid_opts[s][key][0] is bool:
                        d1[s][key] = self.config.getboolean(s, key)
                    elif PyConfig.valid_opts[s][key][0] is int:
                        val = self.config[s][key]
                        if val[:2] == '0x' or val[:2] == '0X':
                            d1[s][key] = int(val[2:], base=16)
                        else:    
                            d1[s][key] = self.config.getint(s, key)
                    elif PyConfig.valid_opts[s][key][0] is bytes:
                        d1[s][key] = bytes(self.config[s][key], 'latin_1')
                    else:
                        d1[s][key] = self.config[s][key]

        return (d1, flag)

    def overwrite_config(self, opts, opts_overwrite):
        opts_ow = opts_overwrite.__dict__

        def conv_config_type(key, opt, vopts):
            if isinstance(vopts[key][0], tuple):
                if opt not in vopts[key][0]:
                    PyConfExcept('Unknown option value: ' + opt)
            elif vopts[key][0] is bool:
                opt = bool(opt)
            elif vopts[key][0] is int: 
                if opt[:2] == '0x' or opt[:2] == '0X':
                    opt = int(opt[2:], base=16)
                else:    
                    opt = int(opt)
            elif vopts[key][0] is bytes:
                opt = bytes(opt, 'latin_1')

            return opt

        def do_overwrite(sec, vopts):
            for key in sec.keys():
                opt = opts_ow.get(key, None)
                if opt:  sec[key] = conv_config_type(key, opt, vopts)

        # section global
        sec = opts['global']
        vopts = PyConfig.valid_opts['global']
        do_overwrite(sec, vopts)

        # validate the specified interface
        intf = opts['global']['interface']
        if intf not in vopts['interface'][0]:
            raise PyConfExcept('Interface ' + intf + ' is not supported.')

        # section of the interface
        sec = opts.get(intf, None)
        vopts = PyConfig.valid_opts[intf]
        if not sec:
            raise PyConfExcept('Config file does not have section ' + intf + '.  Stopped.')
        else:  do_overwrite(sec, vopts)

        return opts

class PyOpts:
    def __init__(self):
        self.parser = OptionParser()

    def add_options(self):
        self.parser.add_option('-H', '--host', dest='host',
                  help='Remote server address.')
        self.parser.add_option('-I', '--interface', dest='interface', 
                  help='Select the IPMI interface.  Supported interfaces are kcs, lan and lanplus (default).')
        self.parser.add_option('-p', '--port', dest='port', 
                  help='Remote server UDP port to connect to.  Default is 623.')
        self.parser.add_option('-U', '--user', dest='user', 
                  help='Remote server username.')
        self.parser.add_option('-P', '--password', dest='password', 
                  help='Remote server password.')
        self.parser.add_option('-C', '--cipher_suite', dest='cipher_suite', 
                  help='''The remote server authentication, integrity, and encryption algorithms 
to use for IPMI v2.0 RMCP+ connections.  The default value is 3.''')
        self.parser.add_option('-A', '--auth', dest='auth', 
                  help='''Specify an authentication type to use for IPMI 1.5 RMCP connections. 
Supported types are none, md2, md5 (default), password, and oem.''')
        self.parser.add_option('-L', '--priv', dest='priv', 
                  help='Force session privilege level.  The default level is 4 (=Administrator).')
        self.parser.add_option('-f', '--force', action='store_true', dest='force',
                  help='Force to overwrite the config file with the options given from the command line.')

    def add_more_options(self):
        self.parser.add_option('-l', '--lun', dest='lun', 
                  help='Set destination lun for raw commands.')
        self.parser.add_option('-b', '--dest_chnl', dest='dest_chnl', 
                  help='Set destination channel for bridged request.')
        self.parser.add_option('-t', '--target_addr', dest='target_addr', 
                  help='Bridge IPMI requests to the remote target address.')

    def parse_options(self, opt_str):
        args = opt_str.split(' ')
        args = [x.strip() for x in args]

        options, _ = self.parser.parse_args(args)

        return options



#
# Copyright (c) 2020, Hyve Design Solutions Corporation.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are 
# met:
#
# 1. Redistributions of source code must retain the above copyright 
#    notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright 
#    notice, this list of conditions and the following disclaimer in the 
#    documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of Hyve Design Solutions Corporation nor the names 
#    of its contributors may be used to endorse or promote products 
#    derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY HYVE DESIGN SOLUTIONS CORPORATION AND 
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, 
# BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND 
# FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 
# HYVE DESIGN SOLUTIONS CORPORATION OR CONTRIBUTORS BE LIABLE FOR ANY 
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL 
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS 
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
# STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING 
# IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
# POSSIBILITY OF SUCH DAMAGE.
#
import os, struct, socket, select
from . import rmcp
from . rmcp import RMCP_Ping
from . _rmcp_msg import IPMI15_Message, RMCP_Message
from . _crypto import *
from .. util.exception import PyIntfExcept

class IPMI20_Message(IPMI15_Message):
    seq_num = 1

    def __init__(self, cipher, sseq=0, sid=0, pwd='', k1=None, k2=None):
        self.cipher = cipher
        self.k1 = k1
        self.k2 = k2
        
        # Get this seq_num
        if isinstance(self, IPMI20_Message):
            self.seq_num = IPMI20_Message.seq_num
            IPMI20_Message.seq_num += 1
            if IPMI20_Message.seq_num == 64: IPMI20_Message.seq_num = 0

        super(IPMI20_Message, self).__init__(AUTH_RMCPP, sseq, sid, pwd)

    def _pack_lan_header(self, payload, payload_type):
        payload_len = len(payload)

        # Auth Type (1) | Payload Type (1) | Sess ID (4) | Seq # (4) | Payload Len (2)
        hdr = struct.pack('<BB4sLH', self.auth, payload_type, self.sid, self.sseq, 
                          payload_len)

        return hdr

    def _unpack_lan_header(self, rsp):
        # unpack payload
        try:
            # Auth Type (1) = 06h | Payload Type (1) | Sess ID (4) | Seq # (4) | Payload Len (2)
            t1 = struct.unpack('<BBLLH', rsp[:12])
            payload = rsp[12:]
        except:
            raise PyIntfExcept('Invalid RMCP+ header in response.')        

        # unpack IPMI LAN package
        if len(payload) != t1[-1] or len(payload) < 7:
            raise PyIntfExcept('Invalid RMCP+ payload length in response.')        

        if t1[0] != self.auth:
            raise PyIntfExcept('Invalid RMCP+ header in response.')        
        
        return t1, payload

    def pack(self, cmd, payload=None):
        # Pack IPMI Payload
        if payload is None:
            payload = self._pack_lan_payload(cmd)
        if len(payload) > 65535:
            raise PyIntfExcept('Request data exceeds max length.')

        # pack the IPMI LAN 2.0 header
        payload_type = cmd.payload_type
        if self.cipher[2] != RAKP_NONE:
            # bit 7 = 1b: payload is encrypted
            payload_type = 0x80
        if self.cipher[1] != RAKP_NONE:
            # bit 6 = 1b: payload is unauthenticated
            payload_type |= 0x40

        # Encrypt the payload if necessary
        if self.cipher[2] == AES_CBC_128:
            payload = encrypt_aes_128_cbc(self.k2, payload)

        msg = self._pack_lan_header(payload, payload_type) + payload

        # Add Integrity Check if necessary
        if self.cipher[1] != RAKP_NONE:
            rem = (len(msg) + 2) & 3
            if rem != 0: 
                pad_len = 4 - rem
                pad = struct.pack('B', 0xff) * pad_len
            else: 
                pad_len = 0

            tail = struct.pack('BB', pad_len, 7)    # Pad Length, Next Header
            if pad_len != 0:    tail = pad + tail   
            msg += tail         

            if self.cipher[1] != MD5_128:
                msg += cal_inte_check(self.cipher[1], self.k1, msg)
            else:
                msg += cal_inte_check(self.cipher[1], self.pwd, msg)
        
        # pack RMCP header
        return RMCP_Message.pack(self) + msg

    def unpack(self, rsp):
        # unpack RMCP header
        rsp = RMCP_Message.unpack(self, rsp)

        # Integrity Check if necessary
        if self.cipher[1] != RAKP_NONE:
            auth_code_len = get_inte_len(self.cipher[1])
            auth_code = rsp[-auth_code_len:]
            rsp = rsp[:-auth_code_len]
            if self.cipher[1] != MD5_128:
                auth_code_expected = cal_inte_check(self.cipher[1], self.k1, rsp)
            else:
                auth_code_expected = cal_inte_check(self.cipher[1], self.pwd, rsp)

            if auth_code != auth_code_expected:
                raise PyIntfExcept('Integrity Check failed.')

            pad_len = rsp[-2]
            rsp = rsp[:-(pad_len+2)]

        # unpack IPMI 2.0 header
        t1, payload = self._unpack_lan_header(rsp)

        # Decrypt if necessary
        if self.cipher[2] == AES_CBC_128:
            payload = decrypt_aes_128_cbc(self.k2, payload)
                    
        # unpack IPMI LAN package
        if t1[1] & 0x3f == 0:
            return self._unpack_lan_payload(payload)
        else:
            return (t1[1], payload)



#
# Copyright (c) 2020, Hyve Design Solutions Corporation.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are 
# met:
#
# 1. Redistributions of source code must retain the above copyright 
#    notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright 
#    notice, this list of conditions and the following disclaimer in the 
#    documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of Hyve Design Solutions Corporation nor the names 
#    of its contributors may be used to endorse or promote products 
#    derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY HYVE DESIGN SOLUTIONS CORPORATION AND 
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, 
# BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND 
# FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 
# HYVE DESIGN SOLUTIONS CORPORATION OR CONTRIBUTORS BE LIABLE FOR ANY 
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL 
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS 
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
# STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING 
# IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
# POSSIBILITY OF SUCH DAMAGE.
#
import struct
from . import IPMI_Message
from .. util.exception import PyMesgExcept, PyMesgCCExcept

NETFN_APP = 6

class GetDeviceID(IPMI_Message):
    def __init__(self):
        super(GetDeviceID, self).__init__(NETFN_APP, 0x01)
        self.req_data = None

    def unpack(self, rsp):
        cc, ret = rsp[2:]        
        if cc != 0:
            return super(GetDeviceID, self).unpack(rsp)
        if ret is None:
            raise PyMesgExcept('Get Device ID: unexpected empty response data.') 

        if len(ret) == 11:
            return super(GetDeviceID, self).unpack(rsp, '<BBBBBB3sH')
        elif len(ret) == 15:
            return super(GetDeviceID, self).unpack(rsp, '<BBBBBB3sH4s')
        raise PyMesgExcept('Get Device ID: unexpected response data len = {0}'.format(len(ret)))      

class ColdReset(IPMI_Message):
    def __init__(self):
        super(ColdReset, self).__init__(NETFN_APP, 0x02)
        self.req_data = None

    def unpack(self, rsp):
        return super(ColdReset, self).unpack(rsp)

class GetSelfTest(IPMI_Message):
    def __init__(self):
        super(GetSelfTest, self).__init__(NETFN_APP, 0x04)
        self.req_data = None

    def unpack(self, rsp):
        return super(GetSelfTest, self).unpack(rsp, 'BB')

class GetDeviceGUID(IPMI_Message):
    def __init__(self):
        super(GetDeviceGUID, self).__init__(NETFN_APP, 0x08)
        self.req_data = None

    def unpack(self, rsp):
        return super(GetDeviceGUID, self).unpack(rsp, '16s')

class ResetWatchdog(IPMI_Message):
    def __init__(self):
        super(ResetWatchdog, self).__init__(NETFN_APP, 0x22)
        self.req_data = None

    def unpack(self, rsp):
        try:
            super(ResetWatchdog, self).unpack(rsp)
            return 0
        except PyMesgCCExcept:
            return rsp[2]   # CC

class SetWatchdog(IPMI_Message):
    def __init__(self, use, act, pre_timeout, exp, init_count):
        super(SetWatchdog, self).__init__(NETFN_APP, 0x24)
        self.req_data = struct.pack('<BBBBH', use, act, pre_timeout, exp, init_count)

    def unpack(self, rsp):
        return super(SetWatchdog, self).unpack(rsp)

class GetWatchdog(IPMI_Message):
    def __init__(self):
        super(GetWatchdog, self).__init__(NETFN_APP, 0x25)
        self.req_data = None

    def unpack(self, rsp):
        return super(GetWatchdog, self).unpack(rsp, '<BBBBHH')

class SetEnables(IPMI_Message):
    def __init__(self, enables):
        super(SetEnables, self).__init__(NETFN_APP, 0x2E)
        self.req_data = struct.pack('B', enables)

    def unpack(self, rsp):
        return super(SetEnables, self).unpack(rsp)

class GetEnables(IPMI_Message):
    def __init__(self):
        super(GetEnables, self).__init__(NETFN_APP, 0x2F)
        self.req_data = None

    def unpack(self, rsp):
        return super(GetEnables, self).unpack(rsp, 'B')

class IPMI_SendMsg(IPMI_Message):
    def __init__(self, dest, inner):
        super(IPMI_SendMsg, self).__init__(NETFN_APP, 0x34)
        self.req_data = struct.pack('B', 0x40 | dest) + inner

    def unpack(self, rsp):
        # rsp = (netfn, cmd, cc, rsp_data)
        return super(IPMI_SendMsg, self).unpack(rsp)

class GetChnlAuthCap(IPMI_Message):
    def __init__(self, priv, chnl=0x0e):
        super(GetChnlAuthCap, self).__init__(NETFN_APP, 0x38)
        self.req_data = struct.pack('BB', 0x80 | (0x0f & chnl), 
                                    priv & 0x0f)

    def unpack(self, rsp):
        return super(GetChnlAuthCap, self).unpack(rsp, 'BBBB3sB')

class GetSessChallenge(IPMI_Message):
    def __init__(self, auth, user=None):
        super(GetSessChallenge, self).__init__(NETFN_APP, 0x39)
        if user is None:
            self.req_data = struct.pack('B16s', auth, 0)
        else:
            self.req_data = struct.pack('B16s', auth, user)

    def unpack(self, rsp):
        return super(GetSessChallenge, self).unpack(rsp, '4s16s')

class ActivateSess(IPMI_Message):
    def __init__(self, auth, priv, chg_data, ioseq_num):
        super(ActivateSess, self).__init__(NETFN_APP, 0x3A)
        self.req_data = struct.pack('<BB16sL', auth, priv, chg_data, ioseq_num)

    def unpack(self, rsp):
        return super(ActivateSess, self).unpack(rsp, '<B4sLB')

class SetSessPriv(IPMI_Message):
    def __init__(self, priv):
        super(SetSessPriv, self).__init__(NETFN_APP, 0x3B)
        self.req_data = struct.pack('B', priv & 0x0f)

    def unpack(self, rsp):
        return super(SetSessPriv, self).unpack(rsp, 'B')

class CloseSess(IPMI_Message):
    def __init__(self, sid):
        super(CloseSess, self).__init__(NETFN_APP, 0x3C)
        self.req_data = struct.pack('4s', sid)

    def unpack(self, rsp):
        return super(CloseSess, self).unpack(rsp)

class SetUserAccess(IPMI_Message):
    def __init__(self, chnl, user, priv):
        super(SetUserAccess, self).__init__(NETFN_APP, 0x43)
        self.req_data = struct.pack('BBB', chnl, user & 0x3f, priv & 0x0f)

    def unpack(self, rsp):
        return super(SetUserAccess, self).unpack(rsp)

class GetUserAccess(IPMI_Message):
    def __init__(self, chnl, user):
        super(GetUserAccess, self).__init__(NETFN_APP, 0x44)
        self.req_data = struct.pack('BB', chnl & 0x0f, user & 0x3f)

    def unpack(self, rsp):
        return super(GetUserAccess, self).unpack(rsp, 'BBBB')

class SetUserName(IPMI_Message):
    def __init__(self, user, name):
        super(SetUserName, self).__init__(NETFN_APP, 0x45)
        self.req_data = struct.pack('B16s', user & 0x3f, name)

    def unpack(self, rsp):
        return super(SetUserName, self).unpack(rsp)

class GetUserName(IPMI_Message):
    def __init__(self, user):
        super(GetUserName, self).__init__(NETFN_APP, 0x46)
        self.req_data = struct.pack('B', user & 0x3f)

    def unpack(self, rsp):
        return super(GetUserName, self).unpack(rsp, '16s')

class SetUserPass(IPMI_Message):
    def __init__(self, user, op, passwd):
        super(SetUserPass, self).__init__(NETFN_APP, 0x47)
        fmt = 'BB20s' if user & 0x80 else 'BB16s'
        self.req_data = struct.pack(fmt, user & 0xbf, op & 3, passwd)

    def unpack(self, rsp):
        try:
            super(SetUserPass, self).unpack(rsp)
        except PyMesgCCExcept as e:
            if e.cc == 0x80:
                print('Password test failed (CC=80h).')
            elif e.cc == 0x81:
                print('Password test failed (CC=81h).')
            else:
                print(e)
        finally:
            return rsp[2]   # CC

class SetSysInfo(IPMI_Message):
    def __init__(self, param, data):
        super(SetSysInfo, self).__init__(NETFN_APP, 0x58)
        self.req_data = struct.pack('B', param) + data

    def unpack(self, rsp):
        try:
            super(SetSysInfo, self).unpack(rsp)
            return 0
        except PyMesgCCExcept:
            return rsp[2]   # CC

class GetSysInfo(IPMI_Message):
    def __init__(self, param, sel, blk):
        super(GetSysInfo, self).__init__(NETFN_APP, 0x59)
        self.req_data = struct.pack('BBBB', 0, param, sel, blk)

    def unpack(self, rsp):
        try:
            super(GetSysInfo, self).unpack(rsp)
            ret = rsp[3]
            return ret[1:]    
        except PyMesgCCExcept:
            return None



#
# Copyright (c) 2020, Hyve Design Solutions Corporation.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are 
# met:
#
# 1. Redistributions of source code must retain the above copyright 
#    notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright 
#    notice, this list of conditions and the following disclaimer in the 
#    documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of Hyve Design Solutions Corporation nor the names 
#    of its contributors may be used to endorse or promote products 
#    derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY HYVE DESIGN SOLUTIONS CORPORATION AND 
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, 
# BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND 
# FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 
# HYVE DESIGN SOLUTIONS CORPORATION OR CONTRIBUTORS BE LIABLE FOR ANY 
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL 
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS 
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
# STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING 
# IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
# POSSIBILITY OF SUCH DAMAGE.
#
from .. mesg.ipmi_storage import GetSELInfo, GetSELAllocInfo, GetSELTime, \
                                ClearSEL, GetSELEntry, RevSEL
from .. util.exception import PyCmdsArgsExcept
from . _common import conv_time, bcd2str, conv_sensor_type, get_sensor_map, \
                    conv_event_reading, do_command, str2int
from . _consts import GENERIC_EVENT_TYPES, SPECIFIC_EVENT_TYPES

def _sel_info(self, argv):
    ver, rec, free, add_ts, erase_ts, op = self.intf.issue_cmd(GetSELInfo)
    self.print('SEL Version               : ' + bcd2str(ver, True))
    self.print('Record Count              : {0}'.format(rec))
    self.print('Free Space                : {0} bytes'.format(free))
    self.print('Most Recent Addition      : ' + conv_time(add_ts))
    self.print('Most Recent Erase         : ' + conv_time(erase_ts))

    get_op_support = lambda mask: 'yes' if op & mask else 'no'

    self.print('SEL Overflow              : ' + get_op_support(0x80))
    self.print('Delete SEL Supported      : ' + get_op_support(8))
    self.print('Partial Add SEL Supported : ' + get_op_support(4))
    self.print('Reserve SEL Supported     : ' + get_op_support(2))
    self.print('SEL Alloc Info Supported  : ' + get_op_support(1))    

    if op & 1:
        t1 = self.intf.issue_cmd(GetSELAllocInfo)
        UC = 'unspecified'

        alloc_units = '{0}'.format(t1[0]) if t1[0] != 0 else UC
        alloc_unit_size = '{0} bytes'.format(t1[1]) if t1[0] != 0 else UC

        self.print('# of Alloc Units          : ' + alloc_units)
        self.print('Alloc Unit Size           : ' + alloc_unit_size)
        self.print('# Free Units              : {0}'.format(t1[2]))
        self.print('Largest Free Blk          : {0}'.format(t1[3]))
        self.print('Max Record Size           : {0}'.format(t1[4]))    

def get_sel_entries(self):
    _, rec, *_  = self.intf.issue_cmd(GetSELInfo)
    if rec == 0:  yield None

    # Get SEL Entry
    next_id, *sel1 = self.intf.issue_cmd(GetSELEntry, b'\0')
    yield sel1
    sel_count = 1

    while next_id != b'\xff\xff':
        if sel_count > rec: break   # already got more SELs than the total # 
        next_id, *sel1 = self.intf.issue_cmd(GetSELEntry, next_id)
        yield sel1
        sel_count += 1

def _conv_sel_event(event, event_reading_type, sensor_type):
    ret = '(unknown)'

    if event_reading_type >= 1 and event_reading_type <= 0x0c:
        # generic
        events = GENERIC_EVENT_TYPES[event_reading_type]
    elif event_reading_type == 0x6f:
        # sensor-specific
        if sensor_type in SPECIFIC_EVENT_TYPES.keys():
            events = SPECIFIC_EVENT_TYPES[sensor_type]
        else:
            return ret
    else:
        return ret

    if event < len(events):
        ret = events[event]

    return ret

def print_sel_list(self, sel_all, opt=1, sensor_map=None):
    if not sel_all:
        self.print('No SEL entries.')
        return

    for sel1 in sel_all:
        rec_id = sel1[0]
        rec_type = sel1[1]
        if rec_type >= 0xe0:  ts = '(OEM Non-Timestamped)'
        else:  ts = conv_time(sel1[2])

        sensor_type = sel1[5]
        sensor_num = sel1[6]
        event_dir = 'Deassertion' if (sel1[7] & 0x80) >> 7 else 'Assertion'
        event_type = sel1[7] & 0x7f
        event = sel1[8] & 0x0f

        sensor_type_str = conv_sensor_type(sel1[5])
        event_str = _conv_sel_event(event, event_type, sensor_type)

        if opt == 1:
            self.print('{0:>4x} | {1} | {2} #{3:02X}h | {4} | {5}'.format(
                  rec_id, ts, sensor_type_str, sensor_num, event_str, event_dir))
        else:
            sensor_name, *_ = sensor_map.get(sensor_num, ('#{0:02X}h'.format(sensor_num), 0, 0))[0]
            if opt == 2:
                self.print('{0:>4x} | {1} | {2}: {3} | {4} | {5}'.format(
                      rec_id, ts, sensor_type_str, sensor_name, event_str, event_dir))
            else:
                gen_id = sel1[3]
                evm_rev = sel1[4]
                event_data = '{0:02x}{1:02x}{2:02x}'.format(sel1[8], sel1[9], sel1[10])
                if rec_type == 2:
                    rec_type_str = 'System Event Record (02h)'
                elif rec_type >= 0xc0 and rec_type <= 0xdf:
                    rec_type_str = 'OEM Timestamped ({0:02X}h)'.format(rec_type)
                elif rec_type >= 0xe0:
                    rec_type_str = 'OEM Non-Timestamped ({0:02X}h)'.format(rec_type)
                else:
                    rec_type_str = 'Unknown ({0:02X}h)'.format(rec_type)

                self.print('SEL Record ID              : {0:04x}'.format(rec_id))
                self.print('    Record Type            : ' + rec_type_str)
                self.print('    Timestamp              : ' + ts)
                self.print('    Generator ID           : {0:04x}'.format(gen_id))
                self.print('    EvM Revision           : {0:02x}'.format(evm_rev))
                self.print('    Sensor Type            : ' + sensor_type_str)
                self.print('    Sensor                 : {0} ({1:02X}h)'.format(sensor_name, sensor_num))
                self.print('    Event Type             : ' + conv_event_reading(event_type, True))
                self.print('    Event Direction        : ' + event_dir)
                self.print('    Event Data             : ' + event_data)
                self.print('    Description            : ' + event_str)
                self.print('')

def _sel_list(self, argv):
    if argv[0] == 'elist' or argv[0] == 'vlist':
        sensor_map = get_sensor_map(self)
    else:
        sensor_map = None

    sel_all = get_sel_entries(self)
    opt, print_hdl = SEL_PRINT_HDL[argv[0]]
    print_hdl(self, sel_all, opt, sensor_map)

def _sel_get(self, argv):
    if len(argv) < 2:
        raise PyCmdsArgsExcept(1)

    rec_id = str2int(argv[1])
    if rec_id == -1:
        raise PyCmdsArgsExcept(3, 0, argv[1])
    rec_id = rec_id.to_bytes(2, 'little')

    # Get SEL Entry
    _, *sel1 = self.intf.issue_cmd(GetSELEntry, rec_id)
    sensor_map = get_sensor_map(self)
    opt, print_hdl = SEL_PRINT_HDL[argv[0]]
    print_hdl(self, (sel1,), opt, sensor_map)

def _sel_clear(self, argv):
    rev, = self.intf.issue_cmd(RevSEL)
    prog, = self.intf.issue_cmd(ClearSEL, rev, 0xaa)

    prog_str = 'unknown'
    if prog == 0:  prog_str = 'in progress'
    elif prog == 1: prog_str = 'completed'
    self.print('SEL Erasure Progress: {0}'.format(prog_str))

def _sel_time(self, argv):
    curr, = self.intf.issue_cmd(GetSELTime)
    self.print('Present Timestamp: {0}'.format(conv_time(curr)))
    
def help_sel(self, argv=None, context=0):
    self.print('SEL Commands:')
    self.print('    list | elist | vlist')
    self.print('    get <record_id>')
    self.print('    clear')
    self.print('    time')
    self.print('    info')
    self.print('    help')

SEL_CMDS = {
    'info': _sel_info,
    'list': _sel_list,
    'elist': _sel_list,
    'vlist': _sel_list,
    'get': _sel_get,
    'clear': _sel_clear,
    'time': _sel_time,
    'help': help_sel,    
}

SEL_PRINT_HDL = {
    'list': (1, print_sel_list),
    'elist': (2, print_sel_list),
    'vlist': (4, print_sel_list),
    'get': (4, print_sel_list),
}

def do_sel(self, argv):
    do_command(self, argv, help_sel, SEL_CMDS)



#
# Copyright (c) 2020, Hyve Design Solutions Corporation.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are 
# met:
#
# 1. Redistributions of source code must retain the above copyright 
#    notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright 
#    notice, this list of conditions and the following disclaimer in the 
#    documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of Hyve Design Solutions Corporation nor the names 
#    of its contributors may be used to endorse or promote products 
#    derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY HYVE DESIGN SOLUTIONS CORPORATION AND 
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, 
# BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND 
# FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 
# HYVE DESIGN SOLUTIONS CORPORATION OR CONTRIBUTORS BE LIABLE FOR ANY 
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL 
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS 
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
# STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING 
# IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
# POSSIBILITY OF SUCH DAMAGE.
#
from .. mesg.ipmi_app import SetUserAccess, GetUserAccess, SetUserName, \
                            GetUserName, SetUserPass
from . _common import str2int, get_chnl, do_command
from . _consts import USER_PRIVS
from .. util.exception import PyCmdsArgsExcept

def _user_info(self, argv):
    chnl = get_chnl(argv[1:])
    max_user, enabled_user, fixed_user, _  = self.intf.issue_cmd(GetUserAccess, chnl, 1)

    self.print('Maximum IDs         : {0}'.format(max_user & 0x3f))
    self.print('Enabled User Count  : {0}'.format(enabled_user & 0x3f))
    self.print('Fixed Name Count    : {0}'.format(fixed_user & 0x3f))

def _user_list(self, argv):
    user_enabled = []
    user_acc = []
    user_name = []

    chnl = get_chnl(argv[1:])
    max_user, enabled, _, acc = self.intf.issue_cmd(GetUserAccess, chnl, 1)
    user_enabled.append((enabled & 0xc0) >> 6)
    user_acc.append(acc)
    
    for i in range(2, max_user + 1):
        try:
            _, enabled, _, acc = self.intf.issue_cmd(GetUserAccess, chnl, i)
        except:
            enabled = 0
            acc = 0x0f
        finally:
            user_enabled.append((enabled & 0xc0) >> 6)
            user_acc.append(acc)

    for i in range(1, max_user + 1):
        try:
            name, = self.intf.issue_cmd(GetUserName, i)
            name = name.decode('latin_1')
            name = name.replace('\x00', '\x20')
        except:
            name = ' ' * 16
        finally:
            user_name.append(name)

    self.print('ID | Name             | Enabling | Callin | Link  | IPMI  | Chnl Priv')
    self.print('-' * 73)
    for i in range(1, max_user + 1):
        enabled = ('unspecified', 'enabled', 'disabled', 'reserved')[user_enabled[i - 1]]
        name = user_name[i - 1]
        acc = user_acc[i - 1]
        callin = 'true' if acc & 0x40 == 0 else 'false'
        link = 'true' if acc & 0x20 else 'false'
        msg = 'true' if acc & 0x10 else 'false'
        priv = USER_PRIVS.get(acc & 0x0f, '(Unknown)')
        fmt = '{0:>2} | {1} | {2:8} | {3:6} | {4:5} | {5:5} | {6} '
        self.print(fmt.format(i, name, enabled, callin, link, msg, priv))

def _user_set_name(self, argv):
    user_id = str2int(argv[0])
    if user_id < 0:
        raise PyCmdsArgsExcept(3, 1, argv[0])

    user_name = b'\0' if len(argv) < 2 else argv[1].encode()
    
    self.intf.issue_cmd(SetUserName, user_id, user_name)

def _user_set_password_imp(self, argv, op):    
    user_id = str2int(argv[0])
    if user_id < 0:
        raise PyCmdsArgsExcept(3, 1, argv[0])
    
    passwd = b'\0' if len(argv) < 2 else argv[1].encode()

    pass_len = '20'
    if len(argv) >= 3:
        if argv[2] in ('20', '16'):
            pass_len = argv[2]
        else:
            raise PyCmdsArgsExcept(3, 1, argv[2])
    if pass_len == '20':    user_id |= 0x80

    return self.intf.issue_cmd(SetUserPass, user_id, op, passwd)

def _user_set_password(self, argv):    
    _user_set_password_imp(self, argv, 2)

def _user_test(self, argv):    
    if len(argv) < 2:
        raise PyCmdsArgsExcept(1, 1)

    cc = _user_set_password_imp(self, argv[1:], 3)
    if cc == 0: self.print('Password test OK.')

def _user_set(self, argv):
    if len(argv) < 3:
        raise PyCmdsArgsExcept(1, 1)
    if argv[1] == 'name':
        _user_set_name(self, argv[2:])
    elif argv[1] == 'password':
        _user_set_password(self, argv[2:])
    elif argv[1] == 'priv':
        _user_set_priv(self, argv[2:])
    else: 
        raise PyCmdsArgsExcept(2, 1, argv[1])

def _user_disable(self, argv):
    if len(argv) < 2:
        raise PyCmdsArgsExcept(1, 2)
    _user_set_password_imp(self, argv[1:], 0)

def _user_enable(self, argv):
    if len(argv) < 2:
        raise PyCmdsArgsExcept(1, 2)
    _user_set_password_imp(self, argv[1:], 1)

def _user_set_priv(self, argv):
    if len(argv) < 2:
        raise PyCmdsArgsExcept(1, 2)

    user_id = str2int(argv[0])
    if user_id < 0:
        raise PyCmdsArgsExcept(3, 2, argv[0])

    priv = str2int(argv[1])
    if priv < 1 or (priv > 5 and priv != 0xf):
        raise PyCmdsArgsExcept(3, 2, argv[1])

    chnl = get_chnl(argv[2:])
    if priv != 0x0f:  chnl |= 0x90
    else:  chnl |= 0x80

    self.intf.issue_cmd(SetUserAccess, chnl, user_id, priv)

def help_user(self, argv=None, context=0):
    if context == 0:
        self.print('User Commands:')
        self.print('    info         [<channel number>]')
        self.print('    list         [<channel number>]')
    if context in (0, 1):
        self.print('    set name     <user id> <username>')
        self.print('    set password <user id> [<password> [<16|20>]]')
        self.print('    test         <user id> [<password> [<16|20>]]')
        self.print('    set priv     <user id> <privilege level> [<channel number>]')
        self.print('        Privilege levels:')
        self.print('         * 0x1 - Callback')
        self.print('         * 0x2 - User')
        self.print('         * 0x3 - Operator')
        self.print('         * 0x4 - Administrator')
        self.print('         * 0x5 - OEM Proprietary')
        self.print('         * 0xF - No Access')
    if context in (0, 2):
        self.print('    enable       <user id>')
        self.print('    disable      <user id>')

USER_CMDS = {
    'info': _user_info,
    'list': _user_list,
    'set': _user_set,
    'test': _user_test,
    'disable': _user_disable,
    'enable': _user_enable,
    'help': help_user, 
}

def do_user(self, argv):
    do_command(self, argv, help_user, USER_CMDS)



import ctypes
import fcntl
import platform

__all__ = (
    'IOC',
    'IO',
    'IOR',
    'IOW',
    'IOWR',
)

class _IoctlGeneric(object):
    _IOC_NRBITS = 8
    _IOC_TYPEBITS = 8
    _IOC_SIZEBITS = 14
    _IOC_DIRBITS = 2
    _IOC_NONE = 0
    _IOC_WRITE = 1
    _IOC_READ = 2

    @classmethod
    def ioc(cls, direction, request_type, request_nr, size):
        _IOC_NRSHIFT = 0
        _IOC_TYPESHIFT = _IOC_NRSHIFT + cls._IOC_NRBITS
        _IOC_SIZESHIFT = _IOC_TYPESHIFT + cls._IOC_TYPEBITS
        _IOC_DIRSHIFT = _IOC_SIZESHIFT + cls._IOC_SIZEBITS
        return (
            (direction << _IOC_DIRSHIFT) |
            (request_type << _IOC_TYPESHIFT) |
            (request_nr << _IOC_NRSHIFT) |
            (size << _IOC_SIZESHIFT)
            )

class _IoctlAlpha(_IoctlGeneric):
    _IOC_NRBITS = 8
    _IOC_TYPEBITS = 8
    _IOC_SIZEBITS = 13
    _IOC_DIRBITS = 3
    _IOC_NONE = 1
    _IOC_READ = 2
    _IOC_WRITE = 4

class _IoctlMips(_IoctlGeneric):
    _IOC_SIZEBITS = 13
    _IOC_DIRBITS = 3
    _IOC_NONE = 1
    _IOC_READ = 2
    _IOC_WRITE = 4

class _IoctlParisc(_IoctlGeneric):
    _IOC_NONE = 0
    _IOC_WRITE = 2
    _IOC_READ = 1

class _IoctlPowerPC(_IoctlGeneric):
    _IOC_SIZEBITS = 13
    _IOC_DIRBITS = 3
    _IOC_NONE = 1
    _IOC_READ = 2
    _IOC_WRITE = 4

class _IoctlSparc(_IoctlGeneric):
    _IOC_NRBITS = 8
    _IOC_TYPEBITS = 8
    _IOC_SIZEBITS = 13
    _IOC_DIRBITS = 3
    _IOC_NONE = 1
    _IOC_READ = 2
    _IOC_WRITE = 4

_machine_ioctl_map = {
    'alpha': _IoctlAlpha,
    'mips': _IoctlMips,
    'mips64': _IoctlMips,
    'parisc': _IoctlParisc,
    'parisc64': _IoctlParisc,
    'ppc': _IoctlPowerPC,
    'ppcle': _IoctlPowerPC,
    'ppc64': _IoctlPowerPC,
    'ppc64le': _IoctlPowerPC,
    'sparc': _IoctlSparc,
    'sparc64': _IoctlSparc,
}

def _machine_ioctl_calculator():
    machine = platform.machine()
    return _machine_ioctl_map.get(machine, _IoctlGeneric)

def _ioc_type_size(size):
    if isinstance(size, type) and issubclass(size, ctypes._SimpleCData):
        return ctypes.sizeof(size)
    elif isinstance(size, int):
        return size
    else:
        raise TypeError('Invalid type for size: {size_type}'.format(size_type=size.__class__.__name__))

def _ioc_request_type(request_type):
    if isinstance(request_type, int):
        return request_type
    if isinstance(request_type, str):
        if len(request_type) > 1:
            raise ValueError('request_type string too long.')
        elif len(request_type) == 0:
            raise ValueError('request_type cannot be an empty string.')
        return ord(request_type)
    else:
        raise ValueError('request_type must be an integer or a string, but was: {request_type_type}'
                         .format(request_type_type=request_type.__class__.__name__))

def IOC(direction, request_type, request_nr, size):
    """ Python implementation of the ``_IOC(...)`` macro from Linux.

    This is a portable implementation of the ``_IOC(...)`` macro from Linux.
    It takes a set of parameters, and calculates a ioctl request number based on those parameters.

    :param direction: Direction of data transfer in this ioctl. This can be one of:

                      * ``None``: No data transfer.
                      * ``'r'``: Read-only (input) data.
                      * ``'w'``: Write-only (output) data.
                      * ``'rw'``: Read-write (input and output) data.
    :param request_type: The ioctl request type. This can be specified as either a string ``'R'`` or an integer ``123``.
    :param request_nr: The ioctl request number. This is an integer.
    :param size: The number of data bytes transferred in this ioctl.
    :return: The calculated ioctl request number.
    """

    calc = _machine_ioctl_calculator()

    if direction is None:
        direction = calc._IOC_NONE
    elif direction == 'r':
        direction = calc._IOC_READ
    elif direction == 'w':
        direction = calc._IOC_WRITE
    elif direction == 'rw':
        direction = calc._IOC_READ | calc._IOC_WRITE
    else:
        raise ValueError('direction must be None, \'r\', \'w\' or \'rw\'.')

    request_type = _ioc_request_type(request_type)
    return calc.ioc(direction, request_type, request_nr, size)

def IO(request_type, request_nr):
    """ Python implementation of the ``_IO(...)`` macro from Linux.

    This is a portable implementation of the ``_IO(...)`` macro from Linux.
    The ``_IO(...)`` macro calculates a ioctl request number for ioctl request that do not transfer any data.

    :param request_type: The ioctl request type. This can be specified as either a string ``'R'`` or an integer ``123``.
    :param request_nr: The ioctl request number. This is an integer.
    :return: The calculated ioctl request number.
    """

    calc = _machine_ioctl_calculator()
    request_type = _ioc_request_type(request_type)
    return calc.ioc(calc._IOC_NONE, request_type, request_nr, 0)

def IOR(request_type, request_nr, size):
    """ Python implementation of the ``_IOR(...)`` macro from Linux.

    This is a portable implementation of the ``_IOR(...)`` macro from Linux.
    The `_IOR(...)`` macro calculates a ioctl request number for ioctl request that only pass read-only (input) data.

    :param request_type: The ioctl request type. This can be specified as either a string ``'R'`` or an integer ``123``.
    :param request_nr: The ioctl request number. This is an integer.
    :param size: The size of the associated data. This can either be an integer or a ctypes type.
    :return: The calculated ioctl request number.
    """

    calc = _machine_ioctl_calculator()
    request_type = _ioc_request_type(request_type)
    size = _ioc_type_size(size)
    return calc.ioc(calc._IOC_READ, request_type, request_nr, size)

def IOW(request_type, request_nr, size):
    """ Python implementation of the ``_IOW(...)`` macro from Linux.

    This is a portable implementation of the ``_IOW(...)`` macro from Linux.
    The ``_IOW(...)`` macro calculates a ioctl request number for ioctl request that only pass write-only (output) data.

    :param request_type: The ioctl request type. This can be specified as either a string ``'R'`` or an integer ``123``.
    :param request_nr: The ioctl request number. This is an integer.
    :param size: The size of the associated data. This can either be an integer or a ctypes type.
    :return: The calculated ioctl request number.
    """

    calc = _machine_ioctl_calculator()
    request_type = _ioc_request_type(request_type)
    size = _ioc_type_size(size)
    return calc.ioc(calc._IOC_WRITE, request_type, request_nr, size)

def IOWR(request_type, request_nr, size):
    """ Python implementation of the ``_IOWR(...)`` macro from Linux.

    This is a portable implementation of the ``_IOWR(...)`` macro from Linux.
    The ``_IOWR(...)`` macro calculates a ioctl request number for ioctl request that use the data for both reading (input) and writing (output).

    :param request_type: The ioctl request type. This can be specified as either a string ``'R'`` or an integer ``123``.
    :param request_nr: The ioctl request number. This is an integer.
    :param size: The size of the associated data. This can either be an integer or a ctypes type.
    :return: The calculated ioctl request number.
    """

    calc = _machine_ioctl_calculator()
    request_type = _ioc_request_type(request_type)
    size = _ioc_type_size(size)
    return calc.ioc(calc._IOC_READ|calc._IOC_WRITE, request_type, request_nr, size)



#
# Copyright (c) 2021, Hyve Design Solutions Corporation.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are 
# met:
#
# 1. Redistributions of source code must retain the above copyright 
#    notice, this list of conditions and the following disclaimer.
#
# 2. Redistributions in binary form must reproduce the above copyright 
#    notice, this list of conditions and the following disclaimer in the 
#    documentation and/or other materials provided with the distribution.
#
# 3. Neither the name of Hyve Design Solutions Corporation nor the names 
#    of its contributors may be used to endorse or promote products 
#    derived from this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY HYVE DESIGN SOLUTIONS CORPORATION AND 
# CONTRIBUTORS "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, 
# BUT NOT LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND 
# FITNESS FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL 
# HYVE DESIGN SOLUTIONS CORPORATION OR CONTRIBUTORS BE LIABLE FOR ANY 
# DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL 
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS 
# OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) 
# HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, 
# STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING 
# IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE 
# POSSIBILITY OF SUCH DAMAGE.
#
import ctypes, fcntl, os, select, struct, sys
from . ioctl import IOR, IOWR
from . import Intf
from .. mesg import IPMI_Raw
from .. mesg.ipmi_app import IPMI_SendMsg
from .. util.exception import PyIntfExcept, PyIntfSeqExcept

class ipmi_addr(ctypes.Structure):
    _fields_ = [
            ('addr_type', ctypes.c_int),
            ('channel', ctypes.c_short),
            ('data', ctypes.c_char_p),
    ]

class ipmi_si_addr(ctypes.Structure):
    _fields_ = [
            ('addr_type', ctypes.c_int),
            ('channel', ctypes.c_short),
            ('lun', ctypes.c_ubyte),
    ]

class ipmi_ipmb_addr(ctypes.Structure):
    _fields_ = [
            ('addr_type', ctypes.c_int),
            ('channel', ctypes.c_short),
            ('slave_addr', ctypes.c_ubyte),
            ('lun', ctypes.c_ubyte),
    ]

class ipmi_msg(ctypes.Structure):
    _fields_ = [
            ('netfn', ctypes.c_ubyte),
            ('cmd', ctypes.c_ubyte),
            ('data_len', ctypes.c_ushort),
            ('data', ctypes.c_char_p),
    ]

class ipmi_req(ctypes.Structure):
    _fields_ = [
            ('addr', ctypes.c_void_p),
            ('addr_len', ctypes.c_uint),
            ('msgid', ctypes.c_long),
            ('msg', ipmi_msg),
    ]

class ipmi_recv(ctypes.Structure):
    _fields_ = [
            ('recv_type', ctypes.c_int),
            ('addr', ctypes.c_void_p),
            ('addr_len', ctypes.c_uint),
            ('msgid', ctypes.c_long),
            ('msg', ipmi_msg),
    ]

class KCS(Intf):
    def __init__(self, opts):
        self.fd = None
        self.msgid = 1
        self.my_addr = opts.get('bmc_addr', 0x20)
        self.dev_num = opts.get('dev_num', 0)

    def __del__(self):
        self.close()

    def open(self):
        IPMICTL_SET_GETS_EVENTS_CMD = IOR('i', 16, ctypes.c_int)
        IPMICTL_SET_MY_ADDRESS_CMD = IOR('i', 17, ctypes.c_uint)

        DEV_STRS = ('/dev/ipmi', '/dev/ipmi/', '/dev/ipmidev/')
        ipmi_devs = ['{0}{1}'.format(dev, self.dev_num) for dev in DEV_STRS]

        for dev in ipmi_devs:
            try:
                self.fd = os.open(dev, os.O_RDWR)
                break
            except:
                continue

        if self.fd is None:
            # Couldn't open the IPMI driver
            print('Could not open device at ' + ' or '.join(ipmi_devs) + ': ', end='')
            print('No such file or directory')
            return -1

        receive_events = ctypes.c_int(1)
        try:
            fcntl.ioctl(self.fd, IPMICTL_SET_GETS_EVENTS_CMD, receive_events)
        except:
            print('Could not enable event receiver')
            return -1

        if self.my_addr != 0:
            my_addr = ctypes.c_uint(self.my_addr)
            try:
                fcntl.ioctl(self.fd, IPMICTL_SET_MY_ADDRESS_CMD, my_addr)
            except:                
                print('Could not set IPMB address')
                return -1

        return 0

    def close(self):
        try:
            if self.fd:
                os.close(self.fd)
                self.fd = None
        except:
            pass

    def sendrecv(self, req):
        IPMICTL_SEND_COMMAND = IOR('i', 13, ctypes.sizeof(ipmi_req))
        IPMICTL_RECEIVE_MSG_TRUNC = IOWR('i', 11, ctypes.sizeof(ipmi_recv))

        # send command request
        fcntl.ioctl(self.fd, IPMICTL_SEND_COMMAND, req)

        # wait for response
        while True:
            r, _, x = select.select([self.fd], [], [], 3)
            if x: raise PyIntfExcept('select() exception occurred.')
            if r: 
                # receive response
                IPMI_MAX_ADDR_SIZE = 0x20
                IPMI_BUF_SIZE = 1024

                addr = ipmi_addr(0, 0, ctypes.addressof(ctypes.create_string_buffer(IPMI_MAX_ADDR_SIZE)))
                addr_len = ctypes.sizeof(addr) + IPMI_MAX_ADDR_SIZE - ctypes.sizeof(ctypes.c_char_p)
                data = ctypes.create_string_buffer(IPMI_BUF_SIZE)
                recv = ipmi_recv(0, ctypes.addressof(addr), addr_len, 0, 
                                 ipmi_msg(0, 0, IPMI_BUF_SIZE, ctypes.addressof(data)))

                # receive command response
                fcntl.ioctl(self.fd, IPMICTL_RECEIVE_MSG_TRUNC, recv)

                # check message ID
                if recv.msgid != req.msgid:
                    continue

                # compose the response message
                cc = data.raw[0]
                if recv.msg.data_len > 0:
                    rsp_data = data.raw[1:recv.msg.data_len]
                    rsp = [recv.msg.netfn, recv.msg.cmd, cc, rsp_data]
                else:
                    rsp = [recv.msg.netfn, recv.msg.cmd, cc]

                return rsp

            else:
                # select() times out
                raise PyIntfExcept('Times out.  Host has no response.')

    def gen_msg(self, cmd, bridging=False, dest=0, target=0):        
        IPMI_SYSTEM_INTERFACE_ADDR_TYPE = 0x0c
        IPMI_IPMB_ADDR_TYPE	= 0x01
        IPMI_BMC_CHANNEL = 0x0f

        if not bridging:
            bmc_addr = ipmi_si_addr(IPMI_SYSTEM_INTERFACE_ADDR_TYPE, IPMI_BMC_CHANNEL, cmd.lun)
        else:
            bmc_addr = ipmi_ipmb_addr(IPMI_IPMB_ADDR_TYPE, dest, target, cmd.lun)

        if cmd.req_data:
            req_len = len(cmd.req_data)
            req_data = cmd.req_data
        else:
            req_len = 0
            req_data = bytes([])

        msg = ipmi_msg(cmd.netfn, cmd.cmd, req_len, req_data)
        req = ipmi_req(ctypes.addressof(bmc_addr), ctypes.sizeof(bmc_addr), self.msgid, msg)
        self.msgid += 1
        if self.msgid == 0xffff:    self.msgid = 1

        return req

    def issue_bridging_cmd(self, dest, target, req, lun=0):
        cmd = IPMI_Raw(req, lun)
        msg = self.gen_msg(cmd, True, dest, target)
        rsp = self.sendrecv(msg)

        return cmd.unpack(rsp)

    def issue_cmd(self, cmd_cls, *args):
        cmd = cmd_cls(*args)
        msg = self.gen_msg(cmd)
        rsp = self.sendrecv(msg)

        return cmd.unpack(rsp)

    def issue_raw_cmd(self, req, lun=0):        
        return self.issue_cmd(IPMI_Raw, req, lun)


